"""
Hybrid Vision Transformer (ViT) in TensorFlow

A TensorFlow implement of the Hybrid Vision Transformers as described in:

'An Image Is Worth 16 x 16 Words: Transformers for Image Recognition at Scale'
    - https://arxiv.org/abs/2010.11929

`How to train your ViT? Data, Augmentation, and Regularization in Vision Transformers`
    - https://arxiv.org/abs/2106.10270

These hybrid model definitions depend on code in vision_transformer.py.
They were moved here to keep file sizes sane.

Copyright 2021 Martins Bruveris
Copyright 2021 Ross Wightman
"""

from tfimm.models import register_model

from .vit import ViT, ViTConfig

# Model_registry will add each entrypoint function to this
__all__ = []


@register_model
def vit_tiny_r_s16_p8_224():
    """R+ViT-Ti/S16 w/ 8x8 patch hybrid @ 224 x 224."""
    cfg = ViTConfig(
        name="vit_tiny_r_s16_p8_224",
        url="[timm]",
        input_size=(224, 224),
        patch_layer="hybrid_embeddings",
        patch_nb_blocks=(),
        patch_size=8,
        embed_dim=192,
        nb_blocks=12,
        nb_heads=3,
        crop_pct=0.9,
        first_conv="patch_embed/backbone/conv",
    )
    return ViT, cfg


@register_model
def vit_tiny_r_s16_p8_384():
    """R+ViT-Ti/S16 w/ 8x8 patch hybrid @ 384 x 384."""
    cfg = ViTConfig(
        name="vit_tiny_r_s16_p8_384",
        url="[timm]",
        input_size=(384, 384),
        patch_layer="hybrid_embeddings",
        patch_nb_blocks=(),
        patch_size=8,
        embed_dim=192,
        nb_blocks=12,
        nb_heads=3,
        crop_pct=1.0,
        first_conv="patch_embed/backbone/conv",
    )
    return ViT, cfg


@register_model
def vit_small_r26_s32_224():
    """R26+ViT-S/S32 hybrid."""
    cfg = ViTConfig(
        name="vit_small_r26_s32_224",
        url="[timm]",
        input_size=(224, 224),
        patch_layer="hybrid_embeddings",
        patch_nb_blocks=(2, 2, 2, 2),
        patch_size=1,
        embed_dim=384,
        nb_blocks=12,
        nb_heads=6,
        crop_pct=0.9,
        first_conv="patch_embed/backbone/stem/conv",
    )
    return ViT, cfg


@register_model
def vit_small_r26_s32_384():
    """R26+ViT-S/S32 hybrid."""
    cfg = ViTConfig(
        name="vit_small_r26_s32_384",
        url="[timm]",
        input_size=(384, 384),
        patch_layer="hybrid_embeddings",
        patch_nb_blocks=(2, 2, 2, 2),
        patch_size=1,
        embed_dim=384,
        nb_blocks=12,
        nb_heads=6,
        crop_pct=1.0,
        first_conv="patch_embed/backbone/stem/conv",
    )
    return ViT, cfg


@register_model
def vit_base_r50_s16_384():
    """
    R50+ViT-B/16 hybrid from original paper (https://arxiv.org/abs/2010.11929).
    ImageNet-1k weights fine-tuned from in21k @ 384x384, source
    https://github.com/google-research/vision_transformer.
    """
    cfg = ViTConfig(
        name="vit_base_r50_s16_384",
        url="[timm]",
        input_size=(384, 384),
        patch_layer="hybrid_embeddings",
        patch_nb_blocks=(3, 4, 9),
        patch_size=1,
        embed_dim=768,
        nb_blocks=12,
        nb_heads=12,
        crop_pct=1.0,
        first_conv="patch_embed/backbone/stem/conv",
    )
    return ViT, cfg


@register_model
def vit_large_r50_s32_224():
    """R50+ViT-L/S32 hybrid."""
    cfg = ViTConfig(
        name="vit_large_r50_s32_224",
        url="[timm]",
        input_size=(224, 224),
        patch_layer="hybrid_embeddings",
        patch_nb_blocks=(3, 4, 6, 3),
        patch_size=1,
        embed_dim=1024,
        nb_blocks=24,
        nb_heads=16,
        crop_pct=0.9,
        first_conv="patch_embed/backbone/stem/conv",
    )
    return ViT, cfg


@register_model
def vit_large_r50_s32_384():
    """R50+ViT-L/S32 hybrid."""
    cfg = ViTConfig(
        name="vit_large_r50_s32_384",
        url="[timm]",
        input_size=(384, 384),
        patch_layer="hybrid_embeddings",
        patch_nb_blocks=(3, 4, 6, 3),
        patch_size=1,
        embed_dim=1024,
        nb_blocks=24,
        nb_heads=16,
        crop_pct=1.0,
        first_conv="patch_embed/backbone/stem/conv",
    )
    return ViT, cfg


@register_model
def vit_tiny_r_s16_p8_224_in21k():
    """R+ViT-Ti/S16 w/ 8x8 patch hybrid.  ImageNet-21k."""
    cfg = ViTConfig(
        name="vit_tiny_r_s16_p8_224_in21k",
        url="[timm]",
        nb_classes=21843,
        input_size=(224, 224),
        patch_layer="hybrid_embeddings",
        patch_nb_blocks=(),
        patch_size=8,
        embed_dim=192,
        nb_blocks=12,
        nb_heads=3,
        crop_pct=0.9,
        first_conv="patch_embed/backbone/conv",
    )
    return ViT, cfg


@register_model
def vit_small_r26_s32_224_in21k():
    """R26+ViT-S/S32 hybrid. ImageNet-21k."""
    cfg = ViTConfig(
        name="vit_small_r26_s32_224_in21k",
        url="[timm]",
        nb_classes=21843,
        input_size=(224, 224),
        patch_layer="hybrid_embeddings",
        patch_nb_blocks=(2, 2, 2, 2),
        patch_size=1,
        embed_dim=384,
        nb_blocks=12,
        nb_heads=6,
        crop_pct=0.9,
        first_conv="patch_embed/backbone/stem/conv",
    )
    return ViT, cfg


@register_model
def vit_base_r50_s16_224_in21k():
    """
    R50+ViT-B/16 hybrid model from original paper (https://arxiv.org/abs/2010.11929).
    ImageNet-21k weights @ 224x224, source
    https://github.com/google-research/vision_transformer.
    """
    cfg = ViTConfig(
        name="vit_base_r50_s16_224_in21k",
        url="[timm]",
        nb_classes=21843,
        input_size=(224, 224),
        patch_layer="hybrid_embeddings",
        patch_nb_blocks=(3, 4, 9),
        patch_size=1,
        embed_dim=768,
        nb_blocks=12,
        nb_heads=12,
        representation_size=768,
        crop_pct=0.9,
        first_conv="patch_embed/backbone/stem/conv",
    )
    return ViT, cfg


@register_model
def vit_large_r50_s32_224_in21k():
    """R50+ViT-L/S32 hybrid. ImageNet-21k."""
    cfg = ViTConfig(
        name="vit_large_r50_s32_224_in21k",
        url="[timm]",
        nb_classes=21843,
        input_size=(224, 224),
        patch_layer="hybrid_embeddings",
        patch_nb_blocks=(3, 4, 6, 3),
        patch_size=1,
        embed_dim=1024,
        nb_blocks=24,
        nb_heads=16,
        crop_pct=0.9,
        first_conv="patch_embed/backbone/stem/conv",
    )
    return ViT, cfg



import logging
import time
from dataclasses import dataclass
from pathlib import Path

import tensorflow as tf

try:
    import wandb
except ImportError:
    wandb = None
    logging.info("Could not import `wandb`. Logging to W&B not possible.")

from .registry import cfg_serializable


@dataclass
class TrainerConfig:
    """
    Configuration class for :class:`SingleGPUTrainer`.

    Attributes:
        validation_before_training: If ``True``, we perform validation before the first
            training step. Particularly useful if continuing to train an already
            partially trained model.
        validation_every_it: We always perform validation at the end of each epoch.
            If `validation_every_it` is set, we also perform validation during the
            epoch every given number of steps.

        ckpt_dir: Checkpoints and models are only saved if ``ckpt_dir`` is a non-empty
            string.
        init_ckpt_dir: If ``init_ckpt_dir`` is set, we load the model state from this
            checkpoint at the beginning of training. This parameter can be used to
            manually continue fine-tuning from the end state of a previous training
            run. This can be either the path to a single checkpoint or a directory.
            If ``init_ckpt_dir`` is a directory, we will select the latest checkpoint.
        resume_from_ckpt: This parameter is used to recover from crashed or restarting
            compute nodes. If ``True``, we try to load the latest checkpoint from
            ``ckpt_dir`` at the start of training. Note, that this will override any
            state loaded from `init_ckpt_dir``.
        ckpt_every_it: We always save a checkpoint at the end of each epoch. If
            ``ckpt_every_it`` is set, we also save a checkpoint every given number of
            steps.
        ckpt_to_keep: This parameter controls how many checkpoints we want to store. In
            addition to the last ``ckpt_to_keep`` number of checkpoints, we will also
            keep a checkpoint every 12 hours. The latter is useful for long-running
            training jobs.

        display_loss_every_it: If ``verbose=True``, we log to the training loss every
            given number of iterations to ``logging.info``.
        verbose: If ``True`` we log progress information to `logging.info``. If
            ``False`` we suppress logging outputs.
    """

    # Validation
    validation_before_training: bool = True
    validation_every_it: int = -1

    # Checkpointing
    ckpt_dir: str = ""
    init_ckpt: str = ""
    resume_from_ckpt: bool = True
    ckpt_every_it: int = -1
    ckpt_to_keep: int = 3

    # Display
    display_loss_every_it: int = 1000
    verbose: bool = True


@cfg_serializable
class SingleGPUTrainer:
    """
    Trainer class for single GPU training.
    """

    cfg_class = TrainerConfig

    def __init__(
        self,
        problem,
        train_ds,
        val_ds,
        timekeeping,
        log_wandb: bool,
        cfg: TrainerConfig,
    ):
        """Initialize the trainer object"""
        self.problem = problem
        self.train_ds = train_ds
        self.val_ds = val_ds
        self.timekeeping = timekeeping
        self.log_wandb = log_wandb
        self.cfg = cfg

        # Check that we will be able to log to W&B if needed
        if self.log_wandb and not wandb:
            raise ValueError("Cannot log to W&B as `wandb` could not be imported.")

        # Other training-related ops
        self.epoch = tf.Variable(0)
        self.batch_size = self.timekeeping.batch_size

        # Checkpoint related variables
        self.ckpt = None
        self.ckpt_variables = None
        self.ckpt_manager = None
        self.init_ckpt_saver()

        # Restore model state
        self.load_ckpt()

    ###
    # Training
    ###
    def train(self):
        """Training loop"""
        if self.cfg.verbose:
            logging.info("Starting training...")

        # Run validation first
        if self.cfg.validation_before_training:
            self.validation()

        # Save checkpoint before starting training
        self.save_ckpt()

        duration = tf.keras.metrics.Mean(dtype=tf.float32)  # Time tracker
        first_epoch = int(self.epoch.numpy())

        ds_iter = iter(self.train_ds)
        for epoch_idx in range(first_epoch, self.timekeeping.nb_epochs):
            self.problem.start_epoch()

            it = 0
            samples_seen = 0
            last_time = time.time()
            while True:
                # Start a new epoch, if we have seen enough samples
                if (
                    self.timekeeping.nb_samples_per_epoch != -1
                    and samples_seen >= self.timekeeping.nb_samples_per_epoch
                ):
                    break

                try:
                    data = next(ds_iter)  # Fetch next piece of data
                except StopIteration:
                    ds_iter = iter(self.train_ds)  # Reset iterator for next epoch
                    break

                loss, logs = self.problem.train_step(data, it=it)
                samples_seen += self.batch_size
                duration.update_state(time.time() - last_time)
                last_time = time.time()

                # Print training progress
                if it == 0 or (
                    self.cfg.display_loss_every_it > 0
                    and it % self.cfg.display_loss_every_it == 0
                ):
                    self.print_training_progress(it, loss, duration)
                    duration.reset_states()

                # Run validation
                if (
                    it > 0
                    and self.cfg.validation_every_it > 0
                    and it % self.cfg.validation_every_it == 0
                ):
                    self.validation()

                # Save checkpoint
                if (
                    it > 0
                    and self.cfg.ckpt_every_it > 0
                    and it % self.cfg.ckpt_every_it == 0
                ):
                    self.save_ckpt()

                # Log metrics
                if self.log_wandb:
                    logs.update({"train/epoch": epoch_idx})
                    wandb.log(logs, commit=True)

                it += 1

            self.validation()
            # We increase the epoch counter before saving the checkpoint, because when
            # load the checkpoint we want to continue training from the next epoch
            # onwards.
            self.epoch.assign_add(1)
            self.save_ckpt()

        if self.cfg.verbose:
            logging.info("... done training.")

    ###
    # Validation
    ###
    def validation(self):
        if self.val_ds is None:
            return
        if self.cfg.verbose:
            logging.info("Validation...")

        start_time = time.time()
        metric, logs = self.problem.validation(self.val_ds)
        duration = time.time() - start_time

        self.print_val_progress(metric, duration)

        if self.log_wandb:
            wandb.log(logs, commit=False)

    ###
    # Print progress
    ###
    def print_training_progress(self, it, loss, duration):
        sec_per_step = duration.result().numpy()
        samples_per_step = self.batch_size
        samples_per_sec = samples_per_step / sec_per_step
        status = (
            "Train: "
            + f"Epoch {int(self.epoch.numpy())} : "
            + f"Iter {it} "
            + f"samples/sec={samples_per_sec:.1f}, "
            + f"sec/step={sec_per_step:.3f}, "
            + f"Global loss={loss:.5f}"
        )
        if self.cfg.verbose:
            logging.info(status)

    def print_val_progress(self, metric, duration):
        status = (
            "Val: "
            + f"Epoch {int(self.epoch.numpy())} : "
            + f"duration={duration:.1f}s, "
            + f"Metric={metric:.4f}"
        )
        if self.cfg.verbose:
            logging.info(status)

    ###
    # Checkpoint management
    ###
    def init_ckpt_saver(self):
        """Creates a checkpoint and a checkpoint manager"""
        self.ckpt_variables = {"epoch": self.epoch}
        self.ckpt_variables.update(self.problem.ckpt_variables(model_only=False))
        self.ckpt = tf.train.Checkpoint(**self.ckpt_variables)

        if self.cfg.ckpt_dir:
            self.ckpt_manager = tf.train.CheckpointManager(
                checkpoint=self.ckpt,
                directory=self.cfg.ckpt_dir,
                max_to_keep=self.cfg.ckpt_to_keep,
                keep_checkpoint_every_n_hours=12,
            )

    def save_ckpt(self):
        """Save a model checkpoint"""
        if not self.ckpt_manager:
            return

        if self.cfg.verbose:
            logging.info("Saving checkpoint...")

        ckpt_path = self.ckpt_manager.save()
        if self.cfg.verbose:
            if ckpt_path:
                logging.info(f"... checkpoint saved in {ckpt_path}")
            else:
                logging.info("... checkpoint wasn't saved -- Not sure why.")

        # The problem class might want to save models in some other, more readily
        # usable format.
        self.problem.save_model(self.cfg.ckpt_dir)

        if self.cfg.verbose:
            logging.info("... saving finished.")

    def load_ckpt(self):
        """
        Restore state from a checkpoint. If `resume_from_ckpt=True` and a checkppint
        is present, restore that checkpoint. Otherwise, if `init_ckpt` is set, load
        model weights only from `init_ckpt`. If neither is set, don't restore anything.
        """
        if (
            self.cfg.resume_from_ckpt
            and self.ckpt_manager
            and self.ckpt_manager.latest_checkpoint
        ):
            ckpt_path = self.ckpt_manager.latest_checkpoint
            if self.cfg.verbose:
                logging.info(f"Restoring from checkpoint {ckpt_path}...")
            self.ckpt.restore(ckpt_path).expect_partial()
        elif self.cfg.init_ckpt:
            init_ckpt = Path(self.cfg.init_ckpt)
            if init_ckpt.is_dir():
                ckpt_path = tf.train.latest_checkpoint(str(init_ckpt))
            else:
                ckpt_path = init_ckpt
            if self.cfg.verbose:
                logging.info(f"Restoring model only from checkpoint {ckpt_path}...")

            # We create a separate checkpoint object that restores only the
            # model variables, nothing else
            ckpt_variables = self.problem.ckpt_variables(model_only=True)
            ckpt = tf.train.Checkpoint(**ckpt_variables)
            ckpt.restore(str(ckpt_path)).expect_partial()
        else:
            if self.cfg.verbose:
                logging.info("No weights/ckpt loaded, keep model as is...")

        if self.cfg.verbose:
            logging.info("...model loaded.")



from dataclasses import dataclass

import tensorflow as tf

from tfimm.models import (
    EmbeddingModel,
    create_model,
    create_preprocessing,
    model_config,
)

from .registry import cfg_serializable


@dataclass
class ModelConfig:
    model_name: str
    pretrained: str = ""
    model_path: str = ""
    input_size: tuple = ()
    in_channels: int = -1
    nb_classes: int = -1
    drop_rate: float = 0.0
    drop_path_rate: float = 0.0
    attn_drop_rate: float = 0.0


@cfg_serializable
class ModelFactory:
    """Wrapper around `tfimm` models to be used by the training framework."""

    cfg_class = ModelConfig

    def __init__(self, cfg: ModelConfig):
        self.cfg = cfg

    def __call__(self):
        # fmt: off
        kwargs = {}
        # We only pass arguments to `create_model`, if they take non-default values.
        for arg, default in [
            ("input_size", ()), ("in_channels", -1), ("nb_classes", -1),
            ("drop_rate", 0.0), ("drop_path_rate", 0.0), ("attn_drop_rate", 0.0)
        ]:
            if getattr(self.cfg, arg) != default:
                kwargs[arg] = getattr(self.cfg, arg)
        # fmt: on

        model = create_model(
            self.cfg.model_name,
            pretrained=self.cfg.pretrained,
            model_path=self.cfg.model_path,
            **kwargs,
        )
        preprocessing = create_preprocessing(self.cfg.model_name)
        return model, preprocessing

    @property
    def tf_input_shape(self):
        """
        Returns the input shape to be used for keras.Input layers. Some models, e.g.,
        ResNets, can operate on arbitrary sized inputs, others, e.g., transformers,
        need fixed-sized inputs.
        """
        cfg = model_config(self.cfg.model_name)
        if not getattr(cfg, "interpolate_input", True):
            input_shape = *self.cfg.input_size, self.cfg.in_channels
        else:
            input_shape = (None, None, self.cfg.in_channels)
        return input_shape


@dataclass
class SavedModelConfig:
    # Path to where the model is saved.
    path: str
    # Preprocessing function will return a tensor of this dtype. If not provided we
    # default to using `tf.keras.backend.floatx()`.
    dtype: str = ""
    # Preprocessing will subtract the mean and divide by the standard deviation, i.e.,
    # `img = (img - mean) / std`. With the default values preprocessing remains the
    # identity.
    mean: tuple = (0.0, 0.0, 0.0)
    std: tuple = (1.0, 1.0, 1.0)


@cfg_serializable
class SavedModel:
    """
    Class provides a wrapper to a model saved on disk, either as an h5 file or as a
    saved model.
    """

    cfg_class = SavedModelConfig

    def __init__(self, cfg: SavedModelConfig):
        self.cfg = cfg

    def __call__(self):
        model_path = str(self.cfg.path)
        # First we try loading the model as a `keras.Model`
        try:
            model = tf.keras.models.load_model(model_path)
        except ValueError:
            # If that doesn't work we try the saved model format.
            model = tf.saved_model.load(model_path)
        dtype = self.cfg.dtype or tf.keras.backend.floatx()

        def _preprocess(img: tf.Tensor) -> tf.Tensor:
            img = tf.cast(img, dtype=dtype)
            img = (img - self.cfg.mean) / self.cfg.std
            return img

        return model, _preprocess


@dataclass
class EmbeddingModelConfig:
    backbone_name: str
    backbone_pretrained: str = ""
    backbone_model_path: str = ""
    input_size: tuple = -1
    in_channels: int = -1
    embed_dim: int = 512
    drop_rate: float = 0.0
    drop_path_rate: float = 0.0
    attn_drop_rate: float = 0.0


@cfg_serializable
class EmbeddingModelFactory:
    """
    Creates a model to be used for embedding learning problems such as face recognition.
    """

    cfg_class = EmbeddingModelConfig

    def __init__(self, cfg: EmbeddingModelConfig):
        self.cfg = cfg

    def __call__(self):
        # fmt: off
        kwargs = {}
        # We only pass arguments to `create_model`, if they take non-default values.
        for arg, default in [
            ("input_size", ()), ("in_channels", -1), ("drop_rate", 0.0),
            ("drop_path_rate", 0.0), ("attn_drop_rate", 0.0)
        ]:
            if getattr(self.cfg, arg) != default:
                kwargs[arg] = getattr(self.cfg, arg)
        # fmt: on

        backbone = create_model(
            self.cfg.backbone_name,
            pretrained=self.cfg.backbone_pretrained,
            model_path=self.cfg.backbone_model_path,
            nb_classes=0,
            **kwargs,
        )
        model = EmbeddingModel(backbone=backbone, embed_dim=self.cfg.embed_dim)
        preprocessing = create_preprocessing(self.cfg.backbone_name)
        return model, preprocessing

    @property
    def tf_input_shape(self):
        """
        Returns the input shape to be used for keras.Input layers. Some models, e.g.,
        ResNets, can operate on arbitrary sized inputs, others, e.g., transformers,
        need fixed-sized inputs.
        """
        cfg = model_config(self.cfg.backbone_name)
        if not getattr(cfg, "interpolate_input", True):
            input_shape = *self.cfg.input_size, self.cfg.in_channels
        else:
            input_shape = (None, None, self.cfg.in_channels)
        return input_shape



# Copyright (c) Meta Platforms, Inc. and affiliates.
# All rights reserved.

# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.

from .image_encoder import ImageEncoderViT  # noqa: F401
from .mask_decoder import MaskDecoder  # noqa: F401
from .prompt_encoder import PromptEncoder  # noqa: F401
from .sam import Sam  # noqa: F401
from .transformer import TwoWayTransformer  # noqa: F401



"""
Various normalisation layers
"""
import tensorflow as tf


class Affine(tf.keras.layers.Layer):
    """
    Affine normalisation as used in ResMLP networks.

    For NHWC ``x``, we return ``alpha * x + beta``, where ``alpha``, ``beta`` are C
    tensors.
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def build(self, input_shape):
        self.alpha = self.add_weight(
            shape=(input_shape[-1],),
            initializer="ones",
            trainable=True,
            name="alpha",
        )
        self.beta = self.add_weight(
            shape=(input_shape[-1],),
            initializer="zeros",
            trainable=True,
            name="beta",
        )

    def call(self, x):
        x = self.alpha * x + self.beta
        return x


def group_normalize(x, gamma, beta, nb_groups=None, group_size=None, eps=1e-5):
    """
    Applies group-normalization to NHWC ``x`` (see abs/1803.08494, go/dune-gn).

    This function just does the math, if you want a "layer" that creates the
    necessary variables etc., see ``group_norm`` below. You must either specify a fixed
    number of groups ``nb_groups``, which will automatically select a corresponding
    group size depending on the input's number of channels, or you must specify a
    ``group_size``, which leads to an automatic number of groups depending on the
    input's number of channels.

    Author: Lucas Beyer

    Args:
        x: N..C-tensor, the input to group-normalize. For images, this would be a
            NHWC-tensor, for time-series a NTC, for videos a NHWTC or NTHWC, all of
            them work, as normalization includes everything between N and C. Even just
            NC shape works, as C is grouped and normalized.
        gamma: tensor with C entries, learnable scale after normalization.
        beta: tensor with C entries, learnable bias after normalization.
        nb_groups: int, number of groups to normalize over (divides C).
        group_size: int, size of the groups to normalize over (divides C).
        eps: float, a small additive constant to avoid /sqrt(0).

    Returns:
        Group-normalized `x`, of the same shape and type as `x`.
    """
    assert (
        x.shape.ndims >= 2
    ), "Less than 2-dim Tensor passed to GroupNorm. Something's fishy."

    nb_channels = x.shape[-1]
    assert nb_channels is not None, "Cannot apply GroupNorm on dynamic channels."
    assert (nb_groups is None) != (
        group_size is None
    ), "You must specify exactly one of `num_groups`, `group_size`"

    if group_size is not None:
        nb_groups = nb_channels // group_size

    assert nb_channels % nb_groups == 0, "GroupNorm: {} not divisible by {}".format(
        nb_channels, nb_groups
    )

    orig_shape = tf.shape(x)

    # This shape is NHWGS where G is #groups and S is group-size.
    extra_shape = [nb_groups, nb_channels // nb_groups]
    group_shape = tf.concat([orig_shape[:-1], extra_shape], axis=-1)
    x = tf.reshape(x, group_shape)

    # The dimensions to normalize over: HWS for images, but more generally all
    # dimensions except N (batch, first) and G (cross-groups, next-to-last).
    # So more visually, normdims are the dots in N......G. (note the last one is
    # also a dot, not a full-stop, argh!)
    normdims = list(range(1, x.shape.ndims - 2)) + [x.shape.ndims - 1]
    mean, var = tf.nn.moments(x, normdims, keepdims=True)

    # Interestingly, we don't have a beta/gamma per group, but still one per
    # channel, at least according to the original paper. Reshape such that they
    # broadcast correctly.
    beta = tf.reshape(beta, extra_shape)
    gamma = tf.reshape(gamma, extra_shape)
    x = tf.nn.batch_normalization(x, mean, var, beta, gamma, eps)
    return tf.reshape(x, orig_shape)


class GroupNormalization(tf.keras.layers.Layer):
    """
    A group-norm "layer" (see abs/1803.08494 go/dune-gn).

    This function creates beta/gamma variables in a name_scope, and uses them to apply
    ``group_normalize`` on the input ``x``. You can either specify a fixed number of
    groups ``nb_groups``, which will automatically select a corresponding group size
    depending on the input's number of channels, or you must specify a ``group_size``,
    which leads to an automatic number of groups depending on the input's number of
    channels. If you specify neither, the paper's recommended ``nb_groups=32`` is used.

    Authors: Lucas Beyer, Joan Puigcerver.

    Args:
        nb_groups: int, the number of channel-groups to normalize over.
        group_size: int, size of the groups to normalize over.
        eps: float, a small additive constant to avoid /sqrt(0).
        beta_init: initializer for bias, defaults to zeros.
        gamma_init: initializer for scale, defaults to ones.
        **kwargs: other tf.keras.layers.Layer arguments.
    """

    def __init__(
        self,
        nb_groups=None,
        group_size=None,
        eps=1e-5,
        beta_initializer=tf.zeros_initializer(),
        gamma_initializer=tf.ones_initializer(),
        **kwargs
    ):
        super(GroupNormalization, self).__init__(**kwargs)
        if nb_groups is None and group_size is None:
            nb_groups = 32

        self._nb_groups = nb_groups
        self._group_size = group_size
        self._eps = eps
        self._beta_init = beta_initializer
        self._gamma_init = gamma_initializer

    def build(self, input_size):
        channels = input_size[-1]
        assert channels is not None, "Cannot apply GN on dynamic channels."
        self._gamma = self.add_weight(
            name="gamma",
            shape=(channels,),
            initializer=self._gamma_init,
            dtype=self.dtype,
        )
        self._beta = self.add_weight(
            name="beta",
            shape=(channels,),
            initializer=self._beta_init,
            dtype=self.dtype,
        )
        super(GroupNormalization, self).build(input_size)

    def call(self, x):
        return group_normalize(
            x, self._gamma, self._beta, self._nb_groups, self._group_size, self._eps
        )



import re
from dataclasses import dataclass
from typing import List, Optional, Tuple, Union

import tensorflow as tf

from tfimm.layers import (
    DropPath,
    FanoutInitializer,
    PadConv2D,
    PadDepthwiseConv2D,
    act_layer_factory,
    norm_layer_factory,
)
from tfimm.utils import make_divisible


def create_conv2d(
    kernel_size: Union[int, Tuple[int, int], List],
    nb_experts: Optional[int] = None,
    nb_groups: int = 1,
    depthwise: bool = False,
    **kwargs,
):
    """
    Selects a 2D convolution implementation based on arguments. Creates and returns one
    of Conv2D, DepthwiseConv2D.

    Used extensively by EfficientNet, MobileNetV3 and related networks.
    """
    # We change the default value for use_bias to False.
    kwargs["use_bias"] = kwargs.get("use_bias", False)

    if isinstance(kernel_size, list):
        assert nb_experts is None  # MixNet + CondConv combo not supported currently
        assert nb_groups is None  # MixedConv groups are defined by kernel list
        # We're going to use only lists for defining the MixedConv2d kernel groups,
        # ints, tuples, other iterables will continue to pass to normal conv and specify
        # non-square kernels
        raise NotImplementedError("MixedConv2D not implemented yet...")
        # m = MixedConv2d(in_channels, out_channels, kernel_size, **kwargs)
    else:
        if nb_experts is not None:
            raise NotImplementedError("ConvConv2D not implemented yet...")
            # m = CondConv2d(
            #     in_channels, out_channels, kernel_size, groups=groups, **kwargs
            # )
        elif depthwise:
            # Depthwise convolution
            conv = PadDepthwiseConv2D(
                kernel_size=kernel_size,
                depthwise_initializer=FanoutInitializer(depthwise=True),
                **kwargs,
            )
        else:
            # Regular (group-wise) convolution
            conv = PadConv2D(
                kernel_size=kernel_size,
                groups=nb_groups,
                kernel_initializer=FanoutInitializer(nb_groups=nb_groups),
                **kwargs,
            )
    return conv


@dataclass
class BlockArgs:
    """
    Class holding the arguments for a block (DC/IR/etc.). Arguments come from two
    sources. Some are decoded from the block definition string and others (e.g.,
    padding) are set by the global context. Some parameters, such as number of
    filters are updated by the global multipliers.

    Contains a function to decode arguments from the string notation. E.g.::

        ir_r2_k3_s2_e1_i32_o16_se0.25_noskip

    All args can exist in any order except for the leading string which indicates the
    block type.

    * Block type (ir = InvertedResidual, ds = DepthwiseSeparable,
      dsa = DepthwiseSeparable with pointwise activation, cn = ConvBnAct, etc.)
    * r - Number of repeat blocks
    * k - Kernel size
    * s - Stride (int, no tuple)
    * e - Expansion ratio
    * c - Output channels
    * se - Squeuze & excitation ratio
    * n - Activation function ("re", "r6", "hs", "sw", etc.)
    """

    block_type: str
    nb_repeats: int
    nb_experts: Optional[int]
    filters: int
    # Used to deal with in_channels issue in EdgeResidual blocks
    force_in_channels: Optional[int]
    exp_kernel_size: Tuple[int, int]
    dw_kernel_size: Tuple[int, int]
    pw_kernel_size: Tuple[int, int]
    stride: int
    padding: Optional[str]
    dilation_rate: int
    group_size: Optional[int]
    exp_ratio: float
    pw_act: bool
    use_se: bool
    se_ratio: float
    norm_layer: Optional[str]
    act_layer: Optional[str]
    skip_connection: bool
    drop_path_rate: float

    @staticmethod
    def decode(block_string):
        """Gets a block from a string notation of arguments."""
        assert isinstance(block_string, str)
        ops = block_string.split("_")
        options = {"block_type": ops[0]}
        for op in ops[1:]:
            if op == "noskip":
                options["skip"] = False
            elif op == "skip":
                options["skip"] = True
            elif op.startswith("n"):
                act_dict = {
                    "re": "relu",
                    "r6": "relu6",
                    "hs": "hard_swish",
                    "sw": "swish",
                    "mi": "mish",
                }
                options["n"] = act_dict[op[1:]]
            else:
                splits = re.split(r"(\d.*)", op)
                if len(splits) >= 2:
                    key, value = splits[:2]
                    options[key] = value

        skip = False if options["block_type"] == "dsa" else options.get("skip", True)
        if options["block_type"] != "er":
            exp_kernel_size = BlockArgs._parse_ksize(options.get("a", "1"))
            dw_kernel_size = BlockArgs._parse_ksize(options.get("k"))
        else:
            exp_kernel_size = BlockArgs._parse_ksize(options.get("k"))
            dw_kernel_size = (1, 1)

        return BlockArgs(
            block_type=options["block_type"],
            nb_repeats=int(options.get("r")),
            nb_experts=int(options.get("cc", 0)) or None,
            filters=int(options.get("c")),
            force_in_channels=int(options.get("fc", 0)) or None,
            exp_kernel_size=exp_kernel_size,
            dw_kernel_size=dw_kernel_size,
            pw_kernel_size=BlockArgs._parse_ksize(options.get("p", "1")),
            stride=int(options.get("s")),
            padding=None,
            dilation_rate=1,
            group_size=int(options.get("gs")) if "gs" in options else None,
            exp_ratio=float(options.get("e", 1.0)),
            pw_act=options["block_type"] == "dsa",
            use_se=True,  # Which network doesn't use this?
            se_ratio=float(options.get("se", 0.0)),
            norm_layer=None,
            act_layer=options.get("n", None),
            skip_connection=skip,
            drop_path_rate=0.0,
        )

    @staticmethod
    def _parse_ksize(ss) -> Tuple[int, int]:
        if ss.isdigit():
            return int(ss), int(ss)
        else:
            ss = ss.split(".")
            return int(ss[0]), int(ss[1])

    @property
    def nb_groups(self):
        if not self.group_size:  # 0 or None
            return 1  # normal conv with 1 group
        else:
            # Note: group_size == 1 -> depthwise conv
            assert self.filters % self.group_size == 0
            return self.filters // self.group_size


class SqueezeExcite(tf.keras.layers.Layer):
    """
    Squeeze-and-Excitation layer with specific features for the EfficientNet family.

    Args:
        rd_ratio: Ratio of squeeze reduction
        rd_channels: Sets number of reduction channels explicitely. Overrides the
            rd_ratio paramter.
        act_layer: Activation function post reduction convolution
        gate_layer: Activation function for the attention gate
        force_act_layer: Overrides value for act_layer.
    """

    def __init__(
        self,
        rd_ratio: float = 0.25,
        rd_channels: Optional[int] = None,
        act_layer: str = "relu",
        gate_layer: str = "sigmoid",
        force_act_layer: Optional[str] = None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.rd_ratio = rd_ratio
        self.rd_channels = rd_channels

        act_layer = act_layer_factory(force_act_layer or act_layer)
        gate_layer = act_layer_factory(gate_layer)

        self.conv_reduce = None
        self.act1 = act_layer(name="act1")
        self.conv_expand = None
        self.gate = gate_layer(name="gate")

    def build(self, input_shape):
        channels = input_shape[-1]
        rd_channels = self.rd_channels or round(channels * self.rd_ratio)
        self.conv_reduce = tf.keras.layers.Conv2D(
            filters=rd_channels,
            kernel_size=1,
            use_bias=True,
            kernel_initializer=FanoutInitializer(),
            name="conv_reduce",
        )
        self.conv_expand = tf.keras.layers.Conv2D(
            filters=channels,
            kernel_size=1,
            use_bias=True,
            kernel_initializer=FanoutInitializer(),
            name="conv_expand",
        )

    def call(self, x):
        x_se = tf.reduce_mean(x, axis=(1, 2), keepdims=True)
        x_se = self.conv_reduce(x_se)
        x_se = self.act1(x_se)
        x_se = self.conv_expand(x_se)
        x_se = self.gate(x_se)
        x = x * x_se
        return x


class ConvBnAct(tf.keras.layers.Layer):
    """Conv + Norm Layer + Activation with optional skip connection."""

    def __init__(self, cfg: BlockArgs, **kwargs):
        super().__init__(**kwargs)
        self.cfg = cfg
        # This depends on number of input channels and is set during build phase
        self.skip_connection = None

        norm_layer = norm_layer_factory(cfg.norm_layer)
        act_layer = act_layer_factory(cfg.act_layer)

        self.conv = create_conv2d(
            filters=cfg.filters,
            kernel_size=cfg.dw_kernel_size,
            strides=cfg.stride,
            padding=cfg.padding,
            dilation_rate=cfg.dilation_rate,
            name="conv",
        )
        self.bn1 = norm_layer(name="bn1")
        self.act1 = act_layer()
        self.drop_path = None

    def build(self, input_shape):
        in_channels = input_shape[-1]
        self.skip_connection = (
            self.cfg.stride == 1
            and self.cfg.filters == in_channels
            and self.cfg.skip_connection
        )
        if self.skip_connection:
            self.drop_path = DropPath(drop_prob=self.cfg.drop_path_rate)

    def call(self, x, training: bool = False):
        shortcut = x
        x = self.conv(x)
        x = self.bn1(x, training=training)
        x = self.act1(x)
        if self.skip_connection:
            x = self.drop_path(x, training=training)
            x = x + shortcut
        return x


class DepthwiseSeparableConv(tf.keras.layers.Layer):
    """
    Depthwise separable block. Used for DS convolutions in MobileNet-V1 and in the place
    of IR blocks that have no expansion convolution. This is an alternative to an IR
    block with an optional first pw convolution.
    """

    def __init__(self, cfg: BlockArgs, **kwargs):
        super().__init__(**kwargs)
        self.cfg = cfg
        # This depends on number of input channels and is set during build phase
        self.skip_connection = None

        norm_layer = norm_layer_factory(cfg.norm_layer)
        act_layer = act_layer_factory(cfg.act_layer)

        self.conv_dw = create_conv2d(
            kernel_size=cfg.dw_kernel_size,
            strides=cfg.stride,
            padding=cfg.padding,
            dilation_rate=cfg.dilation_rate,
            depthwise=True,
            name="conv_dw",
        )
        self.bn1 = norm_layer(name="bn1")
        self.act1 = act_layer()
        self.se = (
            SqueezeExcite(rd_ratio=cfg.se_ratio, act_layer=cfg.act_layer, name="se")
            if cfg.use_se and cfg.se_ratio > 0.0
            else None
        )
        self.conv_pw = create_conv2d(
            filters=cfg.filters,
            kernel_size=cfg.pw_kernel_size,
            padding=cfg.padding,
            nb_groups=cfg.nb_groups,
            name="conv_pw",
        )
        self.bn2 = norm_layer(name="bn2")
        self.act2 = act_layer() if cfg.pw_act else None
        self.drop_path = None

    def build(self, input_shape):
        in_channels = input_shape[-1]
        self.skip_connection = (
            self.cfg.stride == 1
            and self.cfg.filters == in_channels
            and self.cfg.skip_connection
        )
        if self.skip_connection:
            self.drop_path = DropPath(drop_prob=self.cfg.drop_path_rate)

    def call(self, x, training: bool = False):
        shortcut = x
        x = self.conv_dw(x)
        x = self.bn1(x, training=training)
        x = self.act1(x)
        if self.se is not None:
            x = self.se(x, training=training)
        x = self.conv_pw(x)
        x = self.bn2(x, training=training)
        if self.act2 is not None:
            x = self.act2(x)
        if self.skip_connection:
            x = self.drop_path(x, training=training)
            x = x + shortcut
        return x


class InvertedResidual(tf.keras.layers.Layer):
    """
    Inverted residual block with optional SE.

    Originally used in MobileNet-V2, this layer is often referred to as "MBConv" (Mobile
    inverted bottleneck convolution) and is also used in

    * MNasNet - https://arxiv.org/abs/1807.11626
    * EfficientNet - https://arxiv.org/abs/1905.11946
    * MobileNet-V2 - https://arxiv.org/abs/1801.04381
    * MobileNet-V3 - https://arxiv.org/abs/1905.02244
    """

    def __init__(self, cfg: BlockArgs, **kwargs):
        super().__init__(**kwargs)
        self.cfg = cfg
        # This depends on number of input channels and is set during build phase
        self.skip_connection = None

        norm_layer = norm_layer_factory(cfg.norm_layer)
        act_layer = act_layer_factory(cfg.act_layer)

        # Point-wise expansion
        self.conv_pw = None
        self.bn1 = norm_layer(name="bn1")
        self.act1 = act_layer()

        # Depth-wise convolution
        self.conv_dw = create_conv2d(
            kernel_size=cfg.dw_kernel_size,
            strides=cfg.stride,
            padding=cfg.padding,
            dilation_rate=cfg.dilation_rate,
            depthwise=True,
            name="conv_dw",
        )
        self.bn2 = norm_layer(name="bn2")
        self.act2 = act_layer()

        # Squeeze-and-excitation
        self.se = (
            SqueezeExcite(rd_ratio=cfg.se_ratio, act_layer=cfg.act_layer, name="se")
            if cfg.use_se and cfg.se_ratio > 0.0
            else None
        )

        # Point-wise linear projection
        self.conv_pwl = create_conv2d(
            filters=cfg.filters,
            kernel_size=cfg.pw_kernel_size,
            padding=cfg.padding,
            name="conv_pwl",
        )
        self.bn3 = norm_layer(name="bn3")
        self.drop_path = None

    def build(self, input_shape):
        in_channels = input_shape[-1]
        self.skip_connection = (
            self.cfg.stride == 1
            and self.cfg.filters == in_channels
            and self.cfg.skip_connection
        )
        self.conv_pw = create_conv2d(
            filters=make_divisible(in_channels * self.cfg.exp_ratio, 8),
            kernel_size=self.cfg.exp_kernel_size,
            padding=self.cfg.padding,
            nb_groups=self.cfg.nb_groups,
            name="conv_pw",
        )
        if self.skip_connection:
            self.drop_path = DropPath(drop_prob=self.cfg.drop_path_rate)

    def call(self, x, training: bool = False):
        shortcut = x
        x = self.conv_pw(x)
        x = self.bn1(x, training=training)
        x = self.act1(x)
        x = self.conv_dw(x)
        x = self.bn2(x, training=training)
        x = self.act2(x)
        if self.se is not None:
            x = self.se(x, training=training)
        x = self.conv_pwl(x)
        x = self.bn3(x, training=training)
        if self.skip_connection:
            x = self.drop_path(x, training=training)
            x = x + shortcut
        return x


class EdgeResidual(tf.keras.layers.Layer):
    """
    Residual block with expansion convolution followed by pointwise-linear w/ stride

    Originally introduced in `EfficientNet-EdgeTPU: Creating Accelerator-Optimized
    Neural Networks with AutoML`
    * https://ai.googleblog.com/2019/08/efficientnet-edgetpu-creating.html

    This layer is also called FusedMBConv in the MobileDet, EfficientNet-X, and
    EfficientNet-V2 papers
    * MobileDet - https://arxiv.org/abs/2004.14525
    * EfficientNet-X - https://arxiv.org/abs/2102.05610
    * EfficientNet-V2 - https://arxiv.org/abs/2104.00298
    """

    def __init__(self, cfg: BlockArgs, **kwargs):
        super().__init__(**kwargs)
        self.cfg = cfg
        # This depends on number of input channels and is set during build phase
        self.skip_connection = None

        norm_layer = norm_layer_factory(cfg.norm_layer)
        act_layer = act_layer_factory(cfg.act_layer)

        # Point-wise expansion
        self.conv_exp = None
        self.bn1 = norm_layer(name="bn1")
        self.act1 = act_layer()

        # Squeeze-and-excitation
        self.se = (
            SqueezeExcite(rd_ratio=cfg.se_ratio, act_layer=cfg.act_layer, name="se")
            if cfg.use_se and cfg.se_ratio > 0.0
            else None
        )

        # Point-wise linear projection
        self.conv_pwl = create_conv2d(
            filters=cfg.filters,
            kernel_size=cfg.pw_kernel_size,
            padding=cfg.padding,
            name="conv_pwl",
        )
        self.bn2 = norm_layer(name="bn2")
        self.drop_path = None

    def build(self, input_shape):
        in_channels = input_shape[-1]
        self.skip_connection = (
            self.cfg.stride == 1
            and self.cfg.filters == in_channels
            and self.cfg.skip_connection
        )
        # If force_in_channels is set, we use that value for the purpose of calculating
        # the number of filters in the convolution
        force_in_channels = self.cfg.force_in_channels or in_channels
        self.conv_exp = create_conv2d(
            filters=make_divisible(force_in_channels * self.cfg.exp_ratio, 8),
            kernel_size=self.cfg.exp_kernel_size,
            strides=self.cfg.stride,
            padding=self.cfg.padding,
            nb_groups=self.cfg.nb_groups,
            name="conv_exp",
        )
        if self.skip_connection:
            self.drop_path = DropPath(drop_prob=self.cfg.drop_path_rate)

    def call(self, x, training: bool = False):
        shortcut = x
        x = self.conv_exp(x)
        x = self.bn1(x, training=training)
        x = self.act1(x)
        if self.se is not None:
            x = self.se(x, training=training)
        x = self.conv_pwl(x)
        x = self.bn2(x, training=training)
        if self.skip_connection:
            x = self.drop_path(x, training=training)
            x = x + shortcut
        return x



"""
Original PVT implementation. Included here to enable unit tests

https://github.com/whai362/PVT
"""
from functools import partial

import torch
import torch.nn as nn
import torch.nn.functional as F
from timm.models.layers import DropPath, to_2tuple, trunc_normal_
from timm.models.registry import register_model
from timm.models.vision_transformer import _cfg

__all__ = ["PyramidVisionTransformer"]


class Mlp(nn.Module):
    def __init__(
        self,
        in_features,
        hidden_features=None,
        out_features=None,
        act_layer=nn.GELU,
        drop=0.0,
    ):
        super().__init__()
        out_features = out_features or in_features
        hidden_features = hidden_features or in_features
        self.fc1 = nn.Linear(in_features, hidden_features)
        self.act = act_layer()
        self.fc2 = nn.Linear(hidden_features, out_features)
        self.drop = nn.Dropout(drop)

    def forward(self, x):
        x = self.fc1(x)
        x = self.act(x)
        x = self.drop(x)
        x = self.fc2(x)
        x = self.drop(x)
        return x


class Attention(nn.Module):
    def __init__(
        self,
        dim,
        num_heads=8,
        qkv_bias=False,
        qk_scale=None,
        attn_drop=0.0,
        proj_drop=0.0,
        sr_ratio=1,
    ):
        super().__init__()
        assert (
            dim % num_heads == 0
        ), f"dim {dim} should be divided by num_heads {num_heads}."

        self.dim = dim
        self.num_heads = num_heads
        head_dim = dim // num_heads
        self.scale = qk_scale or head_dim**-0.5

        self.q = nn.Linear(dim, dim, bias=qkv_bias)
        self.kv = nn.Linear(dim, dim * 2, bias=qkv_bias)
        self.attn_drop = nn.Dropout(attn_drop)
        self.proj = nn.Linear(dim, dim)
        self.proj_drop = nn.Dropout(proj_drop)

        self.sr_ratio = sr_ratio
        if sr_ratio > 1:
            self.sr = nn.Conv2d(dim, dim, kernel_size=sr_ratio, stride=sr_ratio)
            self.norm = nn.LayerNorm(dim)

    def forward(self, x, H, W):
        B, N, C = x.shape
        q = (
            self.q(x)
            .reshape(B, N, self.num_heads, C // self.num_heads)
            .permute(0, 2, 1, 3)
        )

        if self.sr_ratio > 1:
            x_ = x.permute(0, 2, 1).reshape(B, C, H, W)
            x_ = self.sr(x_).reshape(B, C, -1).permute(0, 2, 1)
            x_ = self.norm(x_)
            kv = (
                self.kv(x_)
                .reshape(B, -1, 2, self.num_heads, C // self.num_heads)
                .permute(2, 0, 3, 1, 4)
            )
        else:
            kv = (
                self.kv(x)
                .reshape(B, -1, 2, self.num_heads, C // self.num_heads)
                .permute(2, 0, 3, 1, 4)
            )
        k, v = kv[0], kv[1]

        attn = (q @ k.transpose(-2, -1)) * self.scale
        attn = attn.softmax(dim=-1)
        attn = self.attn_drop(attn)

        x = (attn @ v).transpose(1, 2).reshape(B, N, C)
        x = self.proj(x)
        x = self.proj_drop(x)

        return x


class Block(nn.Module):
    def __init__(
        self,
        dim,
        num_heads,
        mlp_ratio=4.0,
        qkv_bias=False,
        qk_scale=None,
        drop=0.0,
        attn_drop=0.0,
        drop_path=0.0,
        act_layer=nn.GELU,
        norm_layer=nn.LayerNorm,
        sr_ratio=1,
    ):
        super().__init__()
        self.norm1 = norm_layer(dim)
        self.attn = Attention(
            dim,
            num_heads=num_heads,
            qkv_bias=qkv_bias,
            qk_scale=qk_scale,
            attn_drop=attn_drop,
            proj_drop=drop,
            sr_ratio=sr_ratio,
        )
        # NOTE: drop path for stochastic depth, we shall see if this is better than
        # dropout here
        self.drop_path = DropPath(drop_path) if drop_path > 0.0 else nn.Identity()
        self.norm2 = norm_layer(dim)
        mlp_hidden_dim = int(dim * mlp_ratio)
        self.mlp = Mlp(
            in_features=dim,
            hidden_features=mlp_hidden_dim,
            act_layer=act_layer,
            drop=drop,
        )

    def forward(self, x, H, W):
        x = x + self.drop_path(self.attn(self.norm1(x), H, W))
        x = x + self.drop_path(self.mlp(self.norm2(x)))

        return x


class PatchEmbed(nn.Module):
    """Image to Patch Embedding"""

    def __init__(self, img_size=224, patch_size=16, in_chans=3, embed_dim=768):
        super().__init__()
        img_size = to_2tuple(img_size)
        patch_size = to_2tuple(patch_size)

        self.img_size = img_size
        self.patch_size = patch_size
        self.H, self.W = img_size[0] // patch_size[0], img_size[1] // patch_size[1]
        self.num_patches = self.H * self.W
        self.proj = nn.Conv2d(
            in_chans, embed_dim, kernel_size=patch_size, stride=patch_size
        )
        self.norm = nn.LayerNorm(embed_dim)

    def forward(self, x):
        B, C, H, W = x.shape

        x = self.proj(x).flatten(2).transpose(1, 2)
        x = self.norm(x)
        H, W = H // self.patch_size[0], W // self.patch_size[1]

        return x, (H, W)


class PyramidVisionTransformer(nn.Module):
    def __init__(
        self,
        img_size=224,
        patch_size=16,
        in_chans=3,
        num_classes=1000,
        embed_dims=[64, 128, 256, 512],
        num_heads=[1, 2, 4, 8],
        mlp_ratios=[4, 4, 4, 4],
        qkv_bias=False,
        qk_scale=None,
        drop_rate=0.0,
        attn_drop_rate=0.0,
        drop_path_rate=0.0,
        norm_layer=nn.LayerNorm,
        depths=[3, 4, 6, 3],
        sr_ratios=[8, 4, 2, 1],
        num_stages=4,
    ):
        super().__init__()
        self.num_classes = num_classes
        self.depths = depths
        self.num_stages = num_stages

        dpr = [
            x.item() for x in torch.linspace(0, drop_path_rate, sum(depths))
        ]  # stochastic depth decay rule
        cur = 0

        for i in range(num_stages):
            patch_embed = PatchEmbed(
                img_size=img_size if i == 0 else img_size // (2 ** (i + 1)),
                patch_size=patch_size if i == 0 else 2,
                in_chans=in_chans if i == 0 else embed_dims[i - 1],
                embed_dim=embed_dims[i],
            )
            num_patches = (
                patch_embed.num_patches
                if i != num_stages - 1
                else patch_embed.num_patches + 1
            )
            pos_embed = nn.Parameter(torch.zeros(1, num_patches, embed_dims[i]))
            pos_drop = nn.Dropout(p=drop_rate)

            block = nn.ModuleList(
                [
                    Block(
                        dim=embed_dims[i],
                        num_heads=num_heads[i],
                        mlp_ratio=mlp_ratios[i],
                        qkv_bias=qkv_bias,
                        qk_scale=qk_scale,
                        drop=drop_rate,
                        attn_drop=attn_drop_rate,
                        drop_path=dpr[cur + j],
                        norm_layer=norm_layer,
                        sr_ratio=sr_ratios[i],
                    )
                    for j in range(depths[i])
                ]
            )
            cur += depths[i]

            setattr(self, f"patch_embed{i + 1}", patch_embed)
            setattr(self, f"pos_embed{i + 1}", pos_embed)
            setattr(self, f"pos_drop{i + 1}", pos_drop)
            setattr(self, f"block{i + 1}", block)

        self.norm = norm_layer(embed_dims[3])

        # cls_token
        self.cls_token = nn.Parameter(torch.zeros(1, 1, embed_dims[3]))

        # classification head
        self.head = (
            nn.Linear(embed_dims[3], num_classes) if num_classes > 0 else nn.Identity()
        )

        # init weights
        for i in range(num_stages):
            pos_embed = getattr(self, f"pos_embed{i + 1}")
            trunc_normal_(pos_embed, std=0.02)
        trunc_normal_(self.cls_token, std=0.02)
        self.apply(self._init_weights)

    def _init_weights(self, m):
        if isinstance(m, nn.Linear):
            trunc_normal_(m.weight, std=0.02)
            if isinstance(m, nn.Linear) and m.bias is not None:
                nn.init.constant_(m.bias, 0)
        elif isinstance(m, nn.LayerNorm):
            nn.init.constant_(m.bias, 0)
            nn.init.constant_(m.weight, 1.0)

    @torch.jit.ignore
    def no_weight_decay(self):
        # return {'pos_embed', 'cls_token'} # has pos_embed may be better
        return {"cls_token"}

    def get_classifier(self):
        return self.head

    def reset_classifier(self, num_classes, global_pool=""):
        self.num_classes = num_classes
        self.head = (
            nn.Linear(self.embed_dim, num_classes) if num_classes > 0 else nn.Identity()
        )

    def _get_pos_embed(self, pos_embed, patch_embed, H, W):
        if H * W == self.patch_embed1.num_patches:
            return pos_embed
        else:
            return (
                F.interpolate(
                    pos_embed.reshape(1, patch_embed.H, patch_embed.W, -1).permute(
                        0, 3, 1, 2
                    ),
                    size=(H, W),
                    mode="bilinear",
                )
                .reshape(1, -1, H * W)
                .permute(0, 2, 1)
            )

    def forward_features(self, x):
        B = x.shape[0]

        for i in range(self.num_stages):
            patch_embed = getattr(self, f"patch_embed{i + 1}")
            pos_embed = getattr(self, f"pos_embed{i + 1}")
            pos_drop = getattr(self, f"pos_drop{i + 1}")
            block = getattr(self, f"block{i + 1}")
            x, (H, W) = patch_embed(x)

            if i == self.num_stages - 1:
                cls_tokens = self.cls_token.expand(B, -1, -1)
                x = torch.cat((cls_tokens, x), dim=1)
                pos_embed_ = self._get_pos_embed(pos_embed[:, 1:], patch_embed, H, W)
                pos_embed = torch.cat((pos_embed[:, 0:1], pos_embed_), dim=1)
            else:
                pos_embed = self._get_pos_embed(pos_embed, patch_embed, H, W)

            x = pos_drop(x + pos_embed)
            for blk in block:
                x = blk(x, H, W)
            if i != self.num_stages - 1:
                x = x.reshape(B, H, W, -1).permute(0, 3, 1, 2).contiguous()

        x = self.norm(x)

        return x[:, 0]

    def forward(self, x):
        x = self.forward_features(x)
        x = self.head(x)

        return x


def _conv_filter(state_dict, patch_size=16):
    """convert patch embedding weight from manual patchify + linear proj to conv"""
    out_dict = {}
    for k, v in state_dict.items():
        if "patch_embed.proj.weight" in k:
            v = v.reshape((v.shape[0], 3, patch_size, patch_size))
        out_dict[k] = v

    return out_dict


@register_model
def pvt_tiny(pretrained=False, **kwargs):
    kwargs.pop("pretrained_cfg", None)
    model = PyramidVisionTransformer(
        patch_size=4,
        embed_dims=[64, 128, 320, 512],
        num_heads=[1, 2, 5, 8],
        mlp_ratios=[8, 8, 4, 4],
        qkv_bias=True,
        norm_layer=partial(nn.LayerNorm, eps=1e-6),
        depths=[2, 2, 2, 2],
        sr_ratios=[8, 4, 2, 1],
        **kwargs,
    )
    model.default_cfg = _cfg()

    return model


@register_model
def pvt_small(pretrained=False, **kwargs):
    kwargs.pop("pretrained_cfg", None)
    model = PyramidVisionTransformer(
        patch_size=4,
        embed_dims=[64, 128, 320, 512],
        num_heads=[1, 2, 5, 8],
        mlp_ratios=[8, 8, 4, 4],
        qkv_bias=True,
        norm_layer=partial(nn.LayerNorm, eps=1e-6),
        depths=[3, 4, 6, 3],
        sr_ratios=[8, 4, 2, 1],
        **kwargs,
    )
    model.default_cfg = _cfg()

    return model


@register_model
def pvt_medium(pretrained=False, **kwargs):
    kwargs.pop("pretrained_cfg", None)
    model = PyramidVisionTransformer(
        patch_size=4,
        embed_dims=[64, 128, 320, 512],
        num_heads=[1, 2, 5, 8],
        mlp_ratios=[8, 8, 4, 4],
        qkv_bias=True,
        norm_layer=partial(nn.LayerNorm, eps=1e-6),
        depths=[3, 4, 18, 3],
        sr_ratios=[8, 4, 2, 1],
        **kwargs,
    )
    model.default_cfg = _cfg()

    return model


@register_model
def pvt_large(pretrained=False, **kwargs):
    kwargs.pop("pretrained_cfg", None)
    model = PyramidVisionTransformer(
        patch_size=4,
        embed_dims=[64, 128, 320, 512],
        num_heads=[1, 2, 5, 8],
        mlp_ratios=[8, 8, 4, 4],
        qkv_bias=True,
        norm_layer=partial(nn.LayerNorm, eps=1e-6),
        depths=[3, 8, 27, 3],
        sr_ratios=[8, 4, 2, 1],
        **kwargs,
    )
    model.default_cfg = _cfg()

    return model


@register_model
def pvt_huge_v2(pretrained=False, **kwargs):
    kwargs.pop("pretrained_cfg", None)
    model = PyramidVisionTransformer(
        patch_size=4,
        embed_dims=[128, 256, 512, 768],
        num_heads=[2, 4, 8, 12],
        mlp_ratios=[8, 8, 4, 4],
        qkv_bias=True,
        norm_layer=partial(nn.LayerNorm, eps=1e-6),
        depths=[3, 10, 60, 3],
        sr_ratios=[8, 4, 2, 1],
        # drop_rate=0.0, drop_path_rate=0.02)
        **kwargs,
    )
    model.default_cfg = _cfg()

    return model



"""
We provide an implementation and pretrained weights for the VGG models.

Paper: Very Deep Convolutional Networks For Large-Scale Image Recognition.
`[arXiv:1409.1556] <https://arxiv.org/abs/1409.1556>`_.

This code has been ported from the
`timm <https://github.com/rwightman/pytorch-image-models>`_ implementation.
"""
# Copyright 2023 Marting Bruveris
from collections import OrderedDict
from dataclasses import dataclass
from typing import List, Tuple

import tensorflow as tf

from tfimm.layers import ClassifierHead, act_layer_factory, norm_layer_factory
from tfimm.models import ModelConfig, keras_serializable, register_model
from tfimm.utils import IMAGENET_DEFAULT_MEAN, IMAGENET_DEFAULT_STD

# Model registry will add each entrypoint fn to this
__all__ = ["VGGConfig", "VGG"]


@dataclass
class VGGConfig(ModelConfig):
    """
    Configuration class for VGG models.

    Parameters:
        name: Name of the model.
        url: URL for pretrained weights.
        nb_classes: Number of classes for classification head.
        in_channels: Number of input image channels.
        input_size: Input image size (height, width)

        layers: List with number of filters for conv layers and "M" for pooling layers.
        nb_features: Number of features in pre-classification head.
        mlp_ratio: Ratio for expanding nb_features in pre-classification head.
        global_pool: Global pooling layers.

        drop_rate: Dropout rate.

        norm_layer: Normalization layer. See :func:`~norm_layer_factory` for possible
            values.
        act_layer: Activation function. See :func:`~act_layer_factory` for possible
            values.

        crop_pct: Crop percentage for ImageNet evaluation.
        interpolation: Interpolation method for ImageNet evaluation.
        mean: Defines preprocessing function. If ``x`` is an image with pixel values
            in (0, 1), the preprocessing function is ``(x - mean) / std``.
        std: Defines preprpocessing function.

        first_conv: Name of first convolutional layer. Used by
            :func:`~tfimm.create_model` to adapt the number in input channels when
            loading pretrained weights.
        classifier: Name of classifier layer. Used by :func:`~tfimm.create_model` to
            adapt the classifier when loading pretrained weights.
    """

    nb_classes: int = 1000
    in_channels: int = 3
    input_size: Tuple[int, int] = (224, 224)

    layers: Tuple = ()
    nb_features: int = 4096
    mlp_ratio: float = 1.0
    global_pool: str = "avg"
    # Regularization
    drop_rate: float = 0.0
    # Other parameters
    norm_layer: str = ""
    act_layer: str = "relu"
    # Parameters for inference
    crop_pct: float = 0.875
    interpolation: str = "bilinear"
    # Preprocessing
    mean: Tuple[float, float, float] = IMAGENET_DEFAULT_MEAN
    std: Tuple[float, float, float] = IMAGENET_DEFAULT_STD
    # Weight transfer
    first_conv: str = "features/0"
    classifier: str = "head/fc"


class ConvMlp(tf.keras.layers.Layer):
    def __init__(
        self,
        out_channels: int,
        kernel_size: int,
        mlp_ratio: float,
        drop_rate: float,
        act_layer: str,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.out_channels = out_channels
        self.kernel_size = kernel_size
        self.mlp_ratio = mlp_ratio
        self.drop_rate = drop_rate
        self.act_layer = act_layer

        act_layer = act_layer_factory(self.act_layer)

        hidden_dim = int(out_channels * mlp_ratio)
        self.fc1 = tf.keras.layers.Conv2D(
            filters=hidden_dim, kernel_size=kernel_size, use_bias=True, name="fc1"
        )
        self.act1 = act_layer(name="act1")
        self.drop = tf.keras.layers.Dropout(rate=self.drop_rate, name="drop")
        self.fc2 = tf.keras.layers.Conv2D(
            filters=out_channels, kernel_size=1, use_bias=True, name="fc2"
        )
        self.act2 = act_layer(name="act2")

    def call(self, x, training=False):
        x = self.fc1(x)
        x = self.act1(x)
        x = self.drop(x, training=training)
        x = self.fc2(x)
        x = self.act2(x)
        return x


@keras_serializable
class VGG(tf.keras.Model):
    """
    Class implementing a VGG network.

    Paper: Very Deep Convolutional Networks For Large-Scale Image Recognition.
    `[arXiv:1409.1556] <https://arxiv.org/abs/1409.1556>`_.

    Parameters:
        cfg: Configuration class for the model.
        **kwargs: Arguments are passed to ``tf.keras.Model``.
    """

    cfg_class = VGGConfig

    def __init__(self, cfg: VGGConfig, **kwargs):
        kwargs["name"] = kwargs.get("name", cfg.name)
        super().__init__(**kwargs)
        self.cfg = cfg

        norm_layer = norm_layer_factory(cfg.norm_layer)
        act_layer = act_layer_factory(cfg.act_layer)

        self.blocks = []  # We can't call it layers, since it conflicts with keras.
        self.block_names = []
        idx = 0
        layer_idx = 0
        for v in self.cfg.layers:
            if v == "M":
                pool = tf.keras.layers.MaxPool2D(
                    pool_size=2, strides=2, name=f"features/{idx}"
                )
                self.blocks.append(pool)
                self.block_names.append(f"layer_{layer_idx}")
                idx += 1
            else:
                conv = tf.keras.layers.Conv2D(
                    filters=v, kernel_size=3, padding="same", name=f"features/{idx}"
                )
                if cfg.norm_layer != "":
                    norm = norm_layer(name=f"features/{idx + 1}")
                    act = act_layer(name=f"features/{idx + 2}")
                    self.blocks.extend([conv, norm, act])
                    self.block_names.extend([None, None, f"layer_{layer_idx}"])
                    idx += 3
                else:
                    act = act_layer(name=f"features/{idx + 1}")
                    self.blocks.extend([conv, act])
                    self.block_names.extend([None, f"layer_{layer_idx}"])
                    idx += 2
            layer_idx += 1

        self.pre_logits = ConvMlp(
            out_channels=self.cfg.nb_features,
            kernel_size=7,
            mlp_ratio=self.cfg.mlp_ratio,
            drop_rate=self.cfg.drop_rate,
            act_layer=self.cfg.act_layer,
            name="pre_logits",
        )
        self.head = ClassifierHead(
            nb_classes=self.cfg.nb_classes,
            pool_type=self.cfg.global_pool,
            drop_rate=self.cfg.drop_rate,
            name="head",
        )

    @property
    def dummy_inputs(self) -> tf.Tensor:
        """Returns a tensor of the correct shape for inference."""
        return tf.zeros((1, *self.cfg.input_size, self.cfg.in_channels))

    @property
    def feature_names(self) -> List[str]:
        """
        Names of features, returned when calling ``call`` with ``return_features=True``.
        """
        _, features = self(self.dummy_inputs, return_features=True)
        return list(features.keys())

    def forward_features(
        self, x, training: bool = False, return_features: bool = False
    ):
        """
        Forward pass through model, excluding the classifier layer. This function is
        useful if the model is used as input for downstream tasks such as object
        detection.

        Arguments:
             x: Input to model
             training: Training or inference phase?
             return_features: If ``True``, we return not only the model output, but a
                dictionary with intermediate features.

        Returns:
            If ``return_features=True``, we return a tuple ``(y, features)``, where
            ``y`` is the model output and ``features`` is a dictionary with
            intermediate features.

            If ``return_features=False``, we return only ``y``.
        """
        features = OrderedDict()
        for block, block_name in zip(self.blocks, self.block_names):
            x = block(x, training=training)
            if block_name is not None:
                features[block_name] = x
        x = self.pre_logits(x, training=training)
        features["features"] = x

        return (x, features) if return_features else x

    def call(self, x, training: bool = False, return_features: bool = False):
        """
        Forward pass through the full model.

        Arguments:
             x: Input to model
             training: Training or inference phase?
             return_features: If ``True``, we return not only the model output, but a
                dictionary with intermediate features.

        Returns:
            If ``return_features=True``, we return a tuple ``(y, features)``, where
            ``y`` is the model output and ``features`` is a dictionary with
            intermediate features.

            If ``return_features=False``, we return only ``y``.
        """
        features = OrderedDict()
        x = self.forward_features(x, training, return_features)
        if return_features:
            x, features = x

        x = self.head(x, training=training)
        features["logits"] = x
        return (x, features) if return_features else x


@register_model
def vgg11():
    """VGG 11-layer model (configuration "A")."""
    cfg = VGGConfig(
        name="vgg11",
        url="[timm]",
        layers=(64, "M", 128, "M", 256, 256, "M", 512, 512, "M", 512, 512, "M"),
    )
    return VGG, cfg


@register_model
def vgg11_bn():
    """VGG 11-layer model (configuration "A") with batch normalisation."""
    cfg = VGGConfig(
        name="vgg11_bn",
        url="[timm]",
        layers=(64, "M", 128, "M", 256, 256, "M", 512, 512, "M", 512, 512, "M"),
        norm_layer="batch_norm",
    )
    return VGG, cfg


@register_model
def vgg13():
    """VGG 13-layer model (configuration "B")."""
    cfg = VGGConfig(
        name="vgg13",
        url="[timm]",
        layers=(
            64,
            64,
            "M",
            128,
            128,
            "M",
            256,
            256,
            "M",
            512,
            512,
            "M",
            512,
            512,
            "M",
        ),
    )
    return VGG, cfg


@register_model
def vgg13_bn():
    """VGG 13-layer model (configuration "B") with batch normalisation."""
    cfg = VGGConfig(
        name="vgg13_bn",
        url="[timm]",
        layers=(
            64,
            64,
            "M",
            128,
            128,
            "M",
            256,
            256,
            "M",
            512,
            512,
            "M",
            512,
            512,
            "M",
        ),
        norm_layer="batch_norm",
    )
    return VGG, cfg


@register_model
def vgg16():
    """VGG 16-layer model (configuration "D")."""
    cfg = VGGConfig(
        name="vgg16",
        url="[timm]",
        layers=(
            # fmt: off
            64, 64, 'M', 128, 128, 'M', 256, 256, 256, 'M', 512, 512, 512,
            'M', 512, 512, 512, 'M'
            # fmt: on
        ),
    )
    return VGG, cfg


@register_model
def vgg16_bn():
    """VGG 16-layer model (configuration "D") with batch normalisation."""
    cfg = VGGConfig(
        name="vgg16_bn",
        url="[timm]",
        layers=(
            # fmt: off
            64, 64, 'M', 128, 128, 'M', 256, 256, 256, 'M',
            512, 512, 512, 'M', 512, 512, 512, 'M'
            # fmt: on
        ),
        norm_layer="batch_norm",
    )
    return VGG, cfg


@register_model
def vgg19():
    """VGG 19-layer model (configuration "E")."""
    cfg = VGGConfig(
        name="vgg19",
        url="[timm]",
        layers=(
            # fmt: off
            64, 64, 'M', 128, 128, 'M', 256, 256, 256, 256, 'M',
            512, 512, 512, 512, 'M', 512, 512, 512, 512, 'M'
            # fmt: on
        ),
    )
    return VGG, cfg


@register_model
def vgg19_bn():
    """VGG 19-layer model (configuration "E") with batch normalisation."""
    cfg = VGGConfig(
        name="vgg19_bn",
        url="[timm]",
        layers=(
            # fmt: off
            64, 64, 'M', 128, 128, 'M', 256, 256, 256, 256, 'M',
            512, 512, 512, 512, 'M', 512, 512, 512, 512, 'M'
            # fmt: on
        ),
        norm_layer="batch_norm",
    )
    return VGG, cfg



"""
Original PVT implementation. Included here to enable unit tests

https://github.com/whai362/PVT
"""
import math
from functools import partial

import torch
import torch.nn as nn
from timm.models.layers import DropPath, to_2tuple, trunc_normal_
from timm.models.registry import register_model
from timm.models.vision_transformer import _cfg

__all__ = ["PyramidVisionTransformerV2"]


class Mlp(nn.Module):
    def __init__(
        self,
        in_features,
        hidden_features=None,
        out_features=None,
        act_layer=nn.GELU,
        drop=0.0,
        linear=False,
    ):
        super().__init__()
        out_features = out_features or in_features
        hidden_features = hidden_features or in_features
        self.fc1 = nn.Linear(in_features, hidden_features)
        self.dwconv = DWConv(hidden_features)
        self.act = act_layer()
        self.fc2 = nn.Linear(hidden_features, out_features)
        self.drop = nn.Dropout(drop)
        self.linear = linear
        if self.linear:
            self.relu = nn.ReLU(inplace=True)
        self.apply(self._init_weights)

    def _init_weights(self, m):
        if isinstance(m, nn.Linear):
            trunc_normal_(m.weight, std=0.02)
            if isinstance(m, nn.Linear) and m.bias is not None:
                nn.init.constant_(m.bias, 0)
        elif isinstance(m, nn.LayerNorm):
            nn.init.constant_(m.bias, 0)
            nn.init.constant_(m.weight, 1.0)
        elif isinstance(m, nn.Conv2d):
            fan_out = m.kernel_size[0] * m.kernel_size[1] * m.out_channels
            fan_out //= m.groups
            m.weight.data.normal_(0, math.sqrt(2.0 / fan_out))
            if m.bias is not None:
                m.bias.data.zero_()

    def forward(self, x, H, W):
        x = self.fc1(x)
        if self.linear:
            x = self.relu(x)
        x = self.dwconv(x, H, W)
        x = self.act(x)
        x = self.drop(x)
        x = self.fc2(x)
        x = self.drop(x)
        return x


class Attention(nn.Module):
    def __init__(
        self,
        dim,
        num_heads=8,
        qkv_bias=False,
        qk_scale=None,
        attn_drop=0.0,
        proj_drop=0.0,
        sr_ratio=1,
        linear=False,
    ):
        super().__init__()
        assert (
            dim % num_heads == 0
        ), f"dim {dim} should be divided by num_heads {num_heads}."

        self.dim = dim
        self.num_heads = num_heads
        head_dim = dim // num_heads
        self.scale = qk_scale or head_dim**-0.5

        self.q = nn.Linear(dim, dim, bias=qkv_bias)
        self.kv = nn.Linear(dim, dim * 2, bias=qkv_bias)
        self.attn_drop = nn.Dropout(attn_drop)
        self.proj = nn.Linear(dim, dim)
        self.proj_drop = nn.Dropout(proj_drop)

        self.linear = linear
        self.sr_ratio = sr_ratio
        if not linear:
            if sr_ratio > 1:
                self.sr = nn.Conv2d(dim, dim, kernel_size=sr_ratio, stride=sr_ratio)
                self.norm = nn.LayerNorm(dim)
        else:
            self.pool = nn.AdaptiveAvgPool2d(7)
            self.sr = nn.Conv2d(dim, dim, kernel_size=1, stride=1)
            self.norm = nn.LayerNorm(dim)
            self.act = nn.GELU()
        self.apply(self._init_weights)

    def _init_weights(self, m):
        if isinstance(m, nn.Linear):
            trunc_normal_(m.weight, std=0.02)
            if isinstance(m, nn.Linear) and m.bias is not None:
                nn.init.constant_(m.bias, 0)
        elif isinstance(m, nn.LayerNorm):
            nn.init.constant_(m.bias, 0)
            nn.init.constant_(m.weight, 1.0)
        elif isinstance(m, nn.Conv2d):
            fan_out = m.kernel_size[0] * m.kernel_size[1] * m.out_channels
            fan_out //= m.groups
            m.weight.data.normal_(0, math.sqrt(2.0 / fan_out))
            if m.bias is not None:
                m.bias.data.zero_()

    def forward(self, x, H, W):
        B, N, C = x.shape
        q = (
            self.q(x)
            .reshape(B, N, self.num_heads, C // self.num_heads)
            .permute(0, 2, 1, 3)
        )

        if not self.linear:
            if self.sr_ratio > 1:
                x_ = x.permute(0, 2, 1).reshape(B, C, H, W)
                x_ = self.sr(x_).reshape(B, C, -1).permute(0, 2, 1)
                x_ = self.norm(x_)
                kv = (
                    self.kv(x_)
                    .reshape(B, -1, 2, self.num_heads, C // self.num_heads)
                    .permute(2, 0, 3, 1, 4)
                )
            else:
                kv = (
                    self.kv(x)
                    .reshape(B, -1, 2, self.num_heads, C // self.num_heads)
                    .permute(2, 0, 3, 1, 4)
                )
        else:
            x_ = x.permute(0, 2, 1).reshape(B, C, H, W)
            x_ = self.sr(self.pool(x_)).reshape(B, C, -1).permute(0, 2, 1)
            x_ = self.norm(x_)
            x_ = self.act(x_)
            kv = (
                self.kv(x_)
                .reshape(B, -1, 2, self.num_heads, C // self.num_heads)
                .permute(2, 0, 3, 1, 4)
            )
        k, v = kv[0], kv[1]

        attn = (q @ k.transpose(-2, -1)) * self.scale
        attn = attn.softmax(dim=-1)
        attn = self.attn_drop(attn)

        x = (attn @ v).transpose(1, 2).reshape(B, N, C)
        x = self.proj(x)
        x = self.proj_drop(x)

        return x


class Block(nn.Module):
    def __init__(
        self,
        dim,
        num_heads,
        mlp_ratio=4.0,
        qkv_bias=False,
        qk_scale=None,
        drop=0.0,
        attn_drop=0.0,
        drop_path=0.0,
        act_layer=nn.GELU,
        norm_layer=nn.LayerNorm,
        sr_ratio=1,
        linear=False,
    ):
        super().__init__()
        self.norm1 = norm_layer(dim)
        self.attn = Attention(
            dim,
            num_heads=num_heads,
            qkv_bias=qkv_bias,
            qk_scale=qk_scale,
            attn_drop=attn_drop,
            proj_drop=drop,
            sr_ratio=sr_ratio,
            linear=linear,
        )
        # NOTE: drop path for stochastic depth, we shall see if this is better than
        # dropout here
        self.drop_path = DropPath(drop_path) if drop_path > 0.0 else nn.Identity()
        self.norm2 = norm_layer(dim)
        mlp_hidden_dim = int(dim * mlp_ratio)
        self.mlp = Mlp(
            in_features=dim,
            hidden_features=mlp_hidden_dim,
            act_layer=act_layer,
            drop=drop,
            linear=linear,
        )

        self.apply(self._init_weights)

    def _init_weights(self, m):
        if isinstance(m, nn.Linear):
            trunc_normal_(m.weight, std=0.02)
            if isinstance(m, nn.Linear) and m.bias is not None:
                nn.init.constant_(m.bias, 0)
        elif isinstance(m, nn.LayerNorm):
            nn.init.constant_(m.bias, 0)
            nn.init.constant_(m.weight, 1.0)
        elif isinstance(m, nn.Conv2d):
            fan_out = m.kernel_size[0] * m.kernel_size[1] * m.out_channels
            fan_out //= m.groups
            m.weight.data.normal_(0, math.sqrt(2.0 / fan_out))
            if m.bias is not None:
                m.bias.data.zero_()

    def forward(self, x, H, W):
        x = x + self.drop_path(self.attn(self.norm1(x), H, W))
        x = x + self.drop_path(self.mlp(self.norm2(x), H, W))

        return x


class OverlapPatchEmbed(nn.Module):
    """Image to Patch Embedding"""

    def __init__(self, img_size=224, patch_size=7, stride=4, in_chans=3, embed_dim=768):
        super().__init__()

        img_size = to_2tuple(img_size)
        patch_size = to_2tuple(patch_size)

        assert max(patch_size) > stride, "Set larger patch_size than stride"

        self.img_size = img_size
        self.patch_size = patch_size
        self.H, self.W = img_size[0] // stride, img_size[1] // stride
        self.num_patches = self.H * self.W
        self.proj = nn.Conv2d(
            in_chans,
            embed_dim,
            kernel_size=patch_size,
            stride=stride,
            padding=(patch_size[0] // 2, patch_size[1] // 2),
        )
        self.norm = nn.LayerNorm(embed_dim)

        self.apply(self._init_weights)

    def _init_weights(self, m):
        if isinstance(m, nn.Linear):
            trunc_normal_(m.weight, std=0.02)
            if isinstance(m, nn.Linear) and m.bias is not None:
                nn.init.constant_(m.bias, 0)
        elif isinstance(m, nn.LayerNorm):
            nn.init.constant_(m.bias, 0)
            nn.init.constant_(m.weight, 1.0)
        elif isinstance(m, nn.Conv2d):
            fan_out = m.kernel_size[0] * m.kernel_size[1] * m.out_channels
            fan_out //= m.groups
            m.weight.data.normal_(0, math.sqrt(2.0 / fan_out))
            if m.bias is not None:
                m.bias.data.zero_()

    def forward(self, x):
        x = self.proj(x)
        _, _, H, W = x.shape
        x = x.flatten(2).transpose(1, 2)
        x = self.norm(x)

        return x, H, W


class PyramidVisionTransformerV2(nn.Module):
    def __init__(
        self,
        img_size=224,
        patch_size=16,
        in_chans=3,
        num_classes=1000,
        embed_dims=[64, 128, 256, 512],
        num_heads=[1, 2, 4, 8],
        mlp_ratios=[4, 4, 4, 4],
        qkv_bias=False,
        qk_scale=None,
        drop_rate=0.0,
        attn_drop_rate=0.0,
        drop_path_rate=0.0,
        norm_layer=nn.LayerNorm,
        depths=[3, 4, 6, 3],
        sr_ratios=[8, 4, 2, 1],
        num_stages=4,
        linear=False,
    ):
        super().__init__()
        self.num_classes = num_classes
        self.depths = depths
        self.num_stages = num_stages

        dpr = [
            x.item() for x in torch.linspace(0, drop_path_rate, sum(depths))
        ]  # stochastic depth decay rule
        cur = 0

        for i in range(num_stages):
            patch_embed = OverlapPatchEmbed(
                img_size=img_size if i == 0 else img_size // (2 ** (i + 1)),
                patch_size=7 if i == 0 else 3,
                stride=4 if i == 0 else 2,
                in_chans=in_chans if i == 0 else embed_dims[i - 1],
                embed_dim=embed_dims[i],
            )

            block = nn.ModuleList(
                [
                    Block(
                        dim=embed_dims[i],
                        num_heads=num_heads[i],
                        mlp_ratio=mlp_ratios[i],
                        qkv_bias=qkv_bias,
                        qk_scale=qk_scale,
                        drop=drop_rate,
                        attn_drop=attn_drop_rate,
                        drop_path=dpr[cur + j],
                        norm_layer=norm_layer,
                        sr_ratio=sr_ratios[i],
                        linear=linear,
                    )
                    for j in range(depths[i])
                ]
            )
            norm = norm_layer(embed_dims[i])
            cur += depths[i]

            setattr(self, f"patch_embed{i + 1}", patch_embed)
            setattr(self, f"block{i + 1}", block)
            setattr(self, f"norm{i + 1}", norm)

        # classification head
        self.head = (
            nn.Linear(embed_dims[3], num_classes) if num_classes > 0 else nn.Identity()
        )

        self.apply(self._init_weights)

    def _init_weights(self, m):
        if isinstance(m, nn.Linear):
            trunc_normal_(m.weight, std=0.02)
            if isinstance(m, nn.Linear) and m.bias is not None:
                nn.init.constant_(m.bias, 0)
        elif isinstance(m, nn.LayerNorm):
            nn.init.constant_(m.bias, 0)
            nn.init.constant_(m.weight, 1.0)
        elif isinstance(m, nn.Conv2d):
            fan_out = m.kernel_size[0] * m.kernel_size[1] * m.out_channels
            fan_out //= m.groups
            m.weight.data.normal_(0, math.sqrt(2.0 / fan_out))
            if m.bias is not None:
                m.bias.data.zero_()

    def freeze_patch_emb(self):
        self.patch_embed1.requires_grad = False

    @torch.jit.ignore
    def no_weight_decay(self):
        return {
            "pos_embed1",
            "pos_embed2",
            "pos_embed3",
            "pos_embed4",
            "cls_token",
        }  # has pos_embed may be better

    def get_classifier(self):
        return self.head

    def reset_classifier(self, num_classes, global_pool=""):
        self.num_classes = num_classes
        self.head = (
            nn.Linear(self.embed_dim, num_classes) if num_classes > 0 else nn.Identity()
        )

    def forward_features(self, x):
        B = x.shape[0]

        for i in range(self.num_stages):
            patch_embed = getattr(self, f"patch_embed{i + 1}")
            block = getattr(self, f"block{i + 1}")
            norm = getattr(self, f"norm{i + 1}")
            x, H, W = patch_embed(x)
            for blk in block:
                x = blk(x, H, W)
            x = norm(x)
            if i != self.num_stages - 1:
                x = x.reshape(B, H, W, -1).permute(0, 3, 1, 2).contiguous()

        return x.mean(dim=1)

    def forward(self, x):
        x = self.forward_features(x)
        x = self.head(x)

        return x


class DWConv(nn.Module):
    def __init__(self, dim=768):
        super(DWConv, self).__init__()
        self.dwconv = nn.Conv2d(dim, dim, 3, 1, 1, bias=True, groups=dim)

    def forward(self, x, H, W):
        B, N, C = x.shape
        x = x.transpose(1, 2).view(B, C, H, W)
        x = self.dwconv(x)
        x = x.flatten(2).transpose(1, 2)

        return x


def _conv_filter(state_dict, patch_size=16):
    """convert patch embedding weight from manual patchify + linear proj to conv"""
    out_dict = {}
    for k, v in state_dict.items():
        if "patch_embed.proj.weight" in k:
            v = v.reshape((v.shape[0], 3, patch_size, patch_size))
        out_dict[k] = v

    return out_dict


@register_model
def pvt_v2_b0(pretrained=False, **kwargs):
    kwargs.pop("pretrained_cfg", None)
    model = PyramidVisionTransformerV2(
        patch_size=4,
        embed_dims=[32, 64, 160, 256],
        num_heads=[1, 2, 5, 8],
        mlp_ratios=[8, 8, 4, 4],
        qkv_bias=True,
        norm_layer=partial(nn.LayerNorm, eps=1e-6),
        depths=[2, 2, 2, 2],
        sr_ratios=[8, 4, 2, 1],
        **kwargs,
    )
    model.default_cfg = _cfg()

    return model


@register_model
def pvt_v2_b1(pretrained=False, **kwargs):
    kwargs.pop("pretrained_cfg", None)
    model = PyramidVisionTransformerV2(
        patch_size=4,
        embed_dims=[64, 128, 320, 512],
        num_heads=[1, 2, 5, 8],
        mlp_ratios=[8, 8, 4, 4],
        qkv_bias=True,
        norm_layer=partial(nn.LayerNorm, eps=1e-6),
        depths=[2, 2, 2, 2],
        sr_ratios=[8, 4, 2, 1],
        **kwargs,
    )
    model.default_cfg = _cfg()

    return model


@register_model
def pvt_v2_b2(pretrained=False, **kwargs):
    kwargs.pop("pretrained_cfg", None)
    model = PyramidVisionTransformerV2(
        patch_size=4,
        embed_dims=[64, 128, 320, 512],
        num_heads=[1, 2, 5, 8],
        mlp_ratios=[8, 8, 4, 4],
        qkv_bias=True,
        norm_layer=partial(nn.LayerNorm, eps=1e-6),
        depths=[3, 4, 6, 3],
        sr_ratios=[8, 4, 2, 1],
        **kwargs,
    )
    model.default_cfg = _cfg()

    return model


@register_model
def pvt_v2_b3(pretrained=False, **kwargs):
    kwargs.pop("pretrained_cfg", None)
    model = PyramidVisionTransformerV2(
        patch_size=4,
        embed_dims=[64, 128, 320, 512],
        num_heads=[1, 2, 5, 8],
        mlp_ratios=[8, 8, 4, 4],
        qkv_bias=True,
        norm_layer=partial(nn.LayerNorm, eps=1e-6),
        depths=[3, 4, 18, 3],
        sr_ratios=[8, 4, 2, 1],
        **kwargs,
    )
    model.default_cfg = _cfg()

    return model


@register_model
def pvt_v2_b4(pretrained=False, **kwargs):
    kwargs.pop("pretrained_cfg", None)
    model = PyramidVisionTransformerV2(
        patch_size=4,
        embed_dims=[64, 128, 320, 512],
        num_heads=[1, 2, 5, 8],
        mlp_ratios=[8, 8, 4, 4],
        qkv_bias=True,
        norm_layer=partial(nn.LayerNorm, eps=1e-6),
        depths=[3, 8, 27, 3],
        sr_ratios=[8, 4, 2, 1],
        **kwargs,
    )
    model.default_cfg = _cfg()

    return model


@register_model
def pvt_v2_b5(pretrained=False, **kwargs):
    kwargs.pop("pretrained_cfg", None)
    model = PyramidVisionTransformerV2(
        patch_size=4,
        embed_dims=[64, 128, 320, 512],
        num_heads=[1, 2, 5, 8],
        mlp_ratios=[4, 4, 4, 4],
        qkv_bias=True,
        norm_layer=partial(nn.LayerNorm, eps=1e-6),
        depths=[3, 6, 40, 3],
        sr_ratios=[8, 4, 2, 1],
        **kwargs,
    )
    model.default_cfg = _cfg()

    return model


@register_model
def pvt_v2_b2_li(pretrained=False, **kwargs):
    kwargs.pop("pretrained_cfg", None)
    model = PyramidVisionTransformerV2(
        patch_size=4,
        embed_dims=[64, 128, 320, 512],
        num_heads=[1, 2, 5, 8],
        mlp_ratios=[8, 8, 4, 4],
        qkv_bias=True,
        norm_layer=partial(nn.LayerNorm, eps=1e-6),
        depths=[3, 4, 6, 3],
        sr_ratios=[8, 4, 2, 1],
        linear=True,
        **kwargs,
    )
    model.default_cfg = _cfg()

    return model



from tfimm.train import (
    ClassificationConfig,
    ExperimentConfig,
    LRExpDecayConfig,
    ModelConfig,
    OptimizerConfig,
    TFDSConfig,
    Timekeeping,
    TrainerConfig,
    run,
)


def main_with_python_config():
    """Start experiment with a config defined by python code."""
    cfg = ExperimentConfig(
        trainer=TrainerConfig(
            display_loss_every_it=5,
            ckpt_dir="/tmp/exp_cifar10",
        ),
        trainer_class="SingleGPUTrainer",
        timekeeping=Timekeeping(
            nb_epochs=3,
            batch_size=32,
            nb_samples_per_epoch=640,
        ),
        timekeeping_class="Timekeeping",
        problem=ClassificationConfig(
            model=ModelConfig(
                model_name="resnet18",
                pretrained="",
                input_size=(64, 64),
                nb_channels=3,
                nb_classes=10,
            ),
            model_class="ModelFactory",
            optimizer=OptimizerConfig(
                optimizer="sgd",
                lr_schedule_class="LRExponentialDecayFactory",
                lr_schedule=LRExpDecayConfig(
                    lr=0.001,
                    lr_decay_rate=0.8,
                    lr_decay_frequency=1,
                ),
            ),
            optimizer_class="OptimizerFactory",
            binary_loss=False,
            weight_decay=0.01,
            mixed_precision=False,
        ),
        problem_class="ClassificationProblem",
        train_dataset=TFDSConfig(
            dataset_name="cifar10",
            split="train",
            input_size=(64, 64),
            batch_size=32,
            repeat=True,
            shuffle=True,
            nb_samples=-1,
            dtype="float32",
        ),
        train_dataset_class="TFDSWrapper",
        val_dataset=TFDSConfig(
            dataset_name="cifar10",
            split="test",
            input_size=(64, 64),
            batch_size=32,
            repeat=False,
            shuffle=False,
            nb_samples=320,
            dtype="float32",
        ),
        val_dataset_class="TFDSWrapper",
        log_wandb=False,
    )

    run(cfg, parse_args=False)


def main_with_cfg_file():
    """Start experiment with a config defined by a config file."""
    cfg = {"cfg_file": "tfimm/train/examples/config.yaml"}
    run(cfg, parse_args=False)


if __name__ == "__main__":
    main_with_python_config()
    main_with_cfg_file()



import tensorflow as tf

from tfimm.layers.norm import Affine, GroupNormalization


def act_layer_factory(act_layer: str):
    """Returns a function that creates the required activation layer."""
    if act_layer in {"linear", "swish", "relu", "gelu", "sigmoid"}:
        return lambda **kwargs: tf.keras.layers.Activation(act_layer, **kwargs)
    if act_layer == "relu6":
        return lambda **kwargs: tf.keras.layers.ReLU(max_value=6, **kwargs)
    else:
        raise ValueError(f"Unknown activation: {act_layer}.")


def norm_layer_factory(norm_layer: str):
    """Returns a function that creates a normalization layer"""
    if norm_layer == "":
        return lambda **kwargs: tf.keras.layers.Activation("linear", **kwargs)

    elif norm_layer == "batch_norm":
        bn_class = tf.keras.layers.BatchNormalization
        bn_args = {
            "momentum": 0.9,  # We use PyTorch default args here
            "epsilon": 1e-5,
        }
        return lambda **kwargs: bn_class(**bn_args, **kwargs)

    elif norm_layer == "batch_norm_tf":  # Batch norm with TF default for epsilon
        bn_class = tf.keras.layers.BatchNormalization
        bn_args = {
            "momentum": 0.9,
            "epsilon": 1e-3,
        }
        return lambda **kwargs: bn_class(**bn_args, **kwargs)

    elif norm_layer == "layer_norm":
        bn_class = tf.keras.layers.LayerNormalization
        bn_args = {"epsilon": 1e-5}  # We use PyTorch default args here
        return lambda **kwargs: bn_class(**bn_args, **kwargs)

    elif norm_layer == "layer_norm_eps_1e-6":
        bn_class = tf.keras.layers.LayerNormalization
        bn_args = {"epsilon": 1e-6}
        return lambda **kwargs: bn_class(**bn_args, **kwargs)

    elif norm_layer == "affine":
        return Affine

    elif norm_layer == "group_norm":
        return GroupNormalization

    elif norm_layer == "group_norm_1grp":
        # Group normalization with one group. Used by PoolFormer.
        bn_class = GroupNormalization
        bn_args = {"nb_groups": 1}
        return lambda **kwargs: bn_class(**bn_args, **kwargs)

    else:
        raise ValueError(f"Unknown normalization layer: {norm_layer}")



"""
We provide an implementation and pretrained weights for the PoolFormer models.

Paper: PoolFormer: MetaFormer is Actually What You Need for Vision.
`[arXiv:2111.11418] <https://arxiv.org/abs/2111.11418>`_.

Original pytorch code and weights from
`poolformer <https://github.com/sail-sg/poolformer>`_ repository.

The following models are available.

* ``poolformer_s12``
* ``poolformer_s24``
* ``poolformer_s36``
* ``poolformer_m36``
* ``poolformer_m48``
"""
# Copyright 2021 Garena Online Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Copyright 2022 Marting Bruveris
from collections import OrderedDict
from dataclasses import dataclass
from typing import List, Tuple

import numpy as np
import tensorflow as tf

from tfimm.layers import ConvMLP, DropPath, PatchEmbeddings, norm_layer_factory
from tfimm.models import ModelConfig, keras_serializable, register_model
from tfimm.utils import IMAGENET_DEFAULT_MEAN, IMAGENET_DEFAULT_STD

# Model registry will add each entrypoint fn to this
__all__ = ["PoolFormer", "PoolFormerConfig"]


@dataclass
class PoolFormerConfig(ModelConfig):
    """
    Configuration class for PoolFormer models.

    Parameters:
        name: Name of the model.
        url: URL for pretrained weights.
        nb_classes: Number of classes for classification head.
        in_channels: Number of input image channels.
        input_size: Input image size (height, width)

        embed_dim: Feature dimensions at each stage.
        nb_blocks: Number of blocks at each stage.
        mlp_ratio: Ratio of mlp hidden dim to embedding dim

        drop_rate: Dropout rate.
        drop_path_rate: Dropout rate for stochastic depth.

        norm_layer: Normalization layer. See :func:`~norm_layer_factory` for possible
            values.
        act_layer: Activation function. See :func:`~act_layer_factory` for possible
            values.
        init_scale: Inital value for layer scale weights.

        crop_pct: Crop percentage for ImageNet evaluation.
        interpolation: Interpolation method for ImageNet evaluation.
        mean: Defines preprocessing function. If ``x`` is an image with pixel values
            in (0, 1), the preprocessing function is ``(x - mean) / std``.
        std: Defines preprpocessing function.

        first_conv: Name of first convolutional layer. Used by
            :func:`~tfimm.create_model` to adapt the number in input channels when
            loading pretrained weights.
        classifier: Name of classifier layer. Used by :func:`~tfimm.create_model` to
            adapt the classifier when loading pretrained weights.
    """

    nb_classes: int = 1000
    in_channels: int = 3
    input_size: Tuple[int, int] = (224, 224)
    embed_dim: Tuple = (64, 128, 320, 512)
    nb_blocks: Tuple = (2, 2, 6, 2)
    mlp_ratio: Tuple = (4.0, 4.0, 4.0, 4.0)
    # Regularization
    drop_rate: float = 0.0
    drop_path_rate: float = 0.0
    # Other parameters
    norm_layer: str = "group_norm_1grp"
    act_layer: str = "gelu"
    init_scale: float = 1e-5
    # Parameters for inference
    crop_pct: float = 0.95
    interpolation: str = "bicubic"
    # Preprocessing
    mean: Tuple[float, float, float] = IMAGENET_DEFAULT_MEAN
    std: Tuple[float, float, float] = IMAGENET_DEFAULT_STD
    # Weight transfer
    first_conv: str = "patch_embed/proj"
    classifier: str = "head"


def _weight_initializers(seed=42):
    """Function returns initilializers to be used in the model."""
    kernel_initializer = tf.keras.initializers.TruncatedNormal(
        mean=0.0, stddev=0.02, seed=seed
    )
    bias_initializer = tf.keras.initializers.Zeros()
    return kernel_initializer, bias_initializer


class PoolFormerBlock(tf.keras.layers.Layer):
    """
    PoolFormer block.

    Args:
        embed_dim: Number of feature dimensions.
        mlp_ratio: Ratio of MLP hidden dim to embedding dim.
        drop_rate: Dropout rate.
        drop_path_rate: Dropout rate for stochastic depth/
        norm_layer: Normalization layer.
        act_layer: Activation function.
        init_scale: Initial value for layer scale weights.
    """

    def __init__(
        self,
        embed_dim: int,
        mlp_ratio: float,
        drop_rate: float,
        drop_path_rate: float,
        norm_layer: str,
        act_layer: str,
        init_scale: float,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.embed_dim = embed_dim
        self.mlp_ratio = mlp_ratio
        self.drop_rate = drop_rate
        self.drop_path_rate = drop_path_rate
        self.norm_layer = norm_layer
        self.act_layer = act_layer
        self.init_scale = init_scale

        norm_layer = norm_layer_factory(norm_layer)
        kernel_initializer, bias_initializer = _weight_initializers()

        self.norm1 = norm_layer(name="norm1")
        self.pool = tf.keras.layers.AveragePooling2D(
            pool_size=3, strides=1, padding="same"
        )
        self.norm2 = norm_layer(name="norm2")
        self.mlp = ConvMLP(
            hidden_dim=int(embed_dim * mlp_ratio),
            embed_dim=embed_dim,
            drop_rate=drop_rate,
            act_layer=act_layer,
            kernel_initializer=kernel_initializer,
            bias_initializer=bias_initializer,
            name="mlp",
        )
        self.layer_scale_1 = None
        self.layer_scale_2 = None
        self.drop_path = DropPath(drop_prob=drop_path_rate)

    def build(self, input_shape):
        self.layer_scale_1 = self.add_weight(
            shape=(self.embed_dim,),
            initializer=tf.keras.initializers.Constant(value=self.init_scale),
            trainable=True,
            name="layer_scale_1",
        )
        self.layer_scale_2 = self.add_weight(
            shape=(self.embed_dim,),
            initializer=tf.keras.initializers.Constant(value=self.init_scale),
            trainable=True,
            name="layer_scale_2",
        )

    def call(self, x, training=False):
        shortcut = x
        x = self.norm1(x, training=training)
        x = self.pool(x) - x  # Token mixer layer
        x = x * self.layer_scale_1
        x = self.drop_path(x, training=training)
        x = x + shortcut

        shortcut = x
        x = self.norm2(x, training=training)
        x = self.mlp(x, training=training)
        x = x * self.layer_scale_2
        x = self.drop_path(x, training=training)
        x = x + shortcut
        return x


@keras_serializable
class PoolFormer(tf.keras.Model):
    """
    Class implementing a PoolFormer network.

    Paper: PoolFormer: MetaFormer is Actually What You Need for Vision.
    `[arXiv:2111.11418] <https://arxiv.org/abs/2111.11418>`_.

    Parameters:
        cfg: Configuration class for the model.
        **kwargs: Arguments are passed to ``tf.keras.Model``.
    """

    cfg_class = PoolFormerConfig

    def __init__(self, cfg: PoolFormerConfig, **kwargs):
        kwargs["name"] = kwargs.get("name", cfg.name)
        super().__init__(**kwargs)
        self.cfg = cfg
        norm_layer = norm_layer_factory(cfg.norm_layer)
        kernel_initializer, bias_initializer = _weight_initializers()

        self.patch_embed = PatchEmbeddings(
            patch_size=7,
            embed_dim=cfg.embed_dim[0],
            stride=4,
            padding=2,
            flatten=False,
            norm_layer="",
            kernel_initializer=kernel_initializer,
            bias_initializer=bias_initializer,
            name="patch_embed",
        )

        # Stochastic depth
        dpr = np.linspace(0.0, cfg.drop_path_rate, sum(cfg.nb_blocks))
        dpr = np.split(dpr, np.cumsum(cfg.nb_blocks))

        self.blocks = OrderedDict()
        for j in range(len(cfg.nb_blocks)):
            for k in range(cfg.nb_blocks[j]):
                self.blocks[f"stage_{j}/block_{k}"] = PoolFormerBlock(
                    embed_dim=cfg.embed_dim[j],
                    mlp_ratio=cfg.mlp_ratio[j],
                    drop_rate=cfg.drop_rate,
                    drop_path_rate=dpr[j][k],
                    norm_layer=cfg.norm_layer,
                    act_layer=cfg.act_layer,
                    init_scale=cfg.init_scale,
                    name=f"network/{2*j}/{k}",
                )
            if j < len(cfg.nb_blocks) - 1:
                self.blocks[f"stage_{j}/downsample"] = PatchEmbeddings(
                    embed_dim=cfg.embed_dim[j + 1],
                    patch_size=3,
                    stride=2,
                    padding=1,
                    flatten=False,
                    norm_layer="",
                    kernel_initializer=kernel_initializer,
                    bias_initializer=bias_initializer,
                    name=f"network/{2*j+1}",
                )

        # Classifier head
        self.norm = norm_layer(name="norm")
        self.pool = tf.keras.layers.GlobalAveragePooling2D()
        self.head = (
            tf.keras.layers.Dense(units=cfg.nb_classes, name="head")
            if cfg.nb_classes > 0
            else tf.keras.layers.Activation("linear")  # Identity layer
        )

    @property
    def dummy_inputs(self) -> tf.Tensor:
        """Returns a tensor of the correct shape for inference."""
        return tf.zeros((1, *self.cfg.input_size, self.cfg.in_channels))

    @property
    def feature_names(self) -> List[str]:
        """
        Names of features, returned when calling ``call`` with ``return_features=True``.
        """
        _, features = self(self.dummy_inputs, return_features=True)
        return list(features.keys())

    def forward_features(
        self, x, training: bool = False, return_features: bool = False
    ):
        """
        Forward pass through model, excluding the classifier layer. This function is
        useful if the model is used as input for downstream tasks such as object
        detection.

        Arguments:
             x: Input to model
             training: Training or inference phase?
             return_features: If ``True``, we return not only the model output, but a
                dictionary with intermediate features.

        Returns:
            If ``return_features=True``, we return a tuple ``(y, features)``, where
            ``y`` is the model output and ``features`` is a dictionary with
            intermediate features.

            If ``return_features=False``, we return only ``y``.
        """
        features = OrderedDict()
        x = self.patch_embed(x, training=training)
        features["patch_embedding"] = x

        for key, block in self.blocks.items():
            x = block(x, training=training)
            features[key] = x

        x = self.norm(x, training=training)
        features["features_all"] = x
        x = self.pool(x)
        features["features"] = x

        return (x, features) if return_features else x

    def call(self, x, training: bool = False, return_features: bool = False):
        """
        Forward pass through the full model.

        Arguments:
             x: Input to model
             training: Training or inference phase?
             return_features: If ``True``, we return not only the model output, but a
                dictionary with intermediate features.

        Returns:
            If ``return_features=True``, we return a tuple ``(y, features)``, where
            ``y`` is the model output and ``features`` is a dictionary with
            intermediate features.

            If ``return_features=False``, we return only ``y``.
        """
        features = OrderedDict()
        x = self.forward_features(x, training, return_features)
        if return_features:
            x, features = x

        x = self.head(x)
        features["logits"] = x
        return (x, features) if return_features else x


@register_model
def poolformer_s12():
    """PoolFormer-S12 model, Params: 12M."""
    cfg = PoolFormerConfig(
        name="poolformer_s12",
        url="[pytorch]https://github.com/sail-sg/poolformer/releases/download/v1.0/poolformer_s12.pth.tar",  # noqa: E501
        embed_dim=(64, 128, 320, 512),
        nb_blocks=(2, 2, 6, 2),
        crop_pct=0.9,
    )
    return PoolFormer, cfg


@register_model
def poolformer_s24():
    """PoolFormer-S24 model, Params: 21M."""
    cfg = PoolFormerConfig(
        name="poolformer_s24",
        url="[pytorch]https://github.com/sail-sg/poolformer/releases/download/v1.0/poolformer_s24.pth.tar",  # noqa: E501
        embed_dim=(64, 128, 320, 512),
        nb_blocks=(4, 4, 12, 4),
        crop_pct=0.9,
    )
    return PoolFormer, cfg


@register_model
def poolformer_s36():
    """PoolFormer-S36 model, Params: 31M."""
    cfg = PoolFormerConfig(
        name="poolformer_s36",
        url="[pytorch]https://github.com/sail-sg/poolformer/releases/download/v1.0/poolformer_s36.pth.tar",  # noqa: E501
        embed_dim=(64, 128, 320, 512),
        nb_blocks=(6, 6, 18, 6),
        init_scale=1e-6,
        crop_pct=0.9,
    )
    return PoolFormer, cfg


@register_model
def poolformer_m36():
    """PoolFormer-M36 model, Params: 56M."""
    cfg = PoolFormerConfig(
        name="poolformer_m36",
        url="[pytorch]https://github.com/sail-sg/poolformer/releases/download/v1.0/poolformer_m36.pth.tar",  # noqa: E501
        embed_dim=(96, 192, 384, 768),
        nb_blocks=(6, 6, 18, 6),
        init_scale=1e-6,
        crop_pct=0.95,
    )
    return PoolFormer, cfg


@register_model
def poolformer_m48():
    """PoolFormer-M48 model, Params: 73M."""
    cfg = PoolFormerConfig(
        name="poolformer_m48",
        url="[pytorch]https://github.com/sail-sg/poolformer/releases/download/v1.0/poolformer_m48.pth.tar",  # noqa: E501
        embed_dim=(96, 192, 384, 768),
        nb_blocks=(8, 8, 24, 8),
        init_scale=1e-6,
        crop_pct=0.95,
    )
    return PoolFormer, cfg



import logging
from collections import abc
from pathlib import Path
from typing import Iterable, List, Tuple, Union
from urllib.parse import urlparse

try:
    import boto3
except ImportError:
    boto3 = None
    logging.info("Could not import `boto3`. S3 support not available")

# When asked to look for TFRecord files we consider these file suffixes
TFRECORD_SUFFIXES = (".tfrecord", ".tfrecords")


def setup_logging(logging_level=logging.INFO):
    """
    Creates a logger that logs to stdout. Sets this logger as the global default.
    Logging format is
        2020-12-05 21:44:09,908: Message.

    Returns:
        Doesn't return anything. Modifies global logger.
    """
    logging.basicConfig(level=logging_level)
    fmt = logging.Formatter("%(asctime)s: %(message)s", datefmt="%y-%b-%d %H:%M:%S")
    logger = logging.getLogger()
    logger.setLevel(logging.INFO)
    # We want to change the format of the root logger, which is the first one
    # in the logger.handlers list. A bit hacky, but there we go.
    root = logger.handlers[0]
    root.setFormatter(fmt)


def collect_tfrecord_files(data_dir: Union[str, Iterable[str]]) -> List[str]:
    """
    We collect all tfrecord files in the locations specified by data_dir.

    Args:
        data_dir: Where to look for tfrecords. This can be either one file, one
            directory or an iterable of files or directories. If `data_dir` is a
            directory, we recurse through subdirectories. On S3, we look at all subkeys.

    Returns:
        List of tfrecord files, possibly empty.
    """
    return collect_files_with_suffix(data_dir, TFRECORD_SUFFIXES)


def collect_files_with_suffix(
    data_dir: Union[str, List[str]], suffix: Union[str, Tuple[str, ...]]
) -> List[str]:
    """
    We collect all files in the locations specified by data_dir with the given suffixes.

    Args:
        data_dir: Where to look for tfrecords. This can be either one file, one
            directory or an iterable of files or directories. If `data_dir` is a
            directory, we recurse through subdirectories. On S3, we look at all subkeys.
        suffix: Which suffixes are we allowed to match

    Returns:
        List of files, possibly empty.
    """
    matching_files = []
    if data_dir is None:
        return matching_files
    elif isinstance(data_dir, abc.Iterable):
        for d in data_dir:
            matching_files.extend(collect_files_with_suffix(d, suffix))
    else:  # Assume it is either string or Path
        data_dir = str(data_dir)
        if data_dir.startswith("s3://"):
            matching_files = _collect_s3_files(data_dir, suffix)
        else:
            matching_files = _collect_local_files(data_dir, suffix)
    matching_files.sort()
    return list(matching_files)


def _collect_local_files(
    data_dir: str, suffix: Union[str, Tuple[str, ...]]
) -> List[str]:
    data_dir = Path(data_dir)
    if data_dir.suffix in suffix:
        matching_files = [str(data_dir)]
    elif data_dir.is_dir():
        matching_files = [str(f) for s in suffix for f in data_dir.rglob(f"*{s}")]
    else:
        matching_files = []
    return matching_files


def _collect_s3_files(data_dir: str, suffix: Union[str, Tuple[str, ...]]) -> List[str]:
    if data_dir.endswith(suffix):
        return [data_dir]

    bucket, prefix = _split_s3_url(data_dir)
    tfrecord_keys = _get_matching_s3_keys(bucket, prefix=prefix, suffix=suffix)
    tfrecord_files = [f"s3://{bucket}/{key}" for key in tfrecord_keys]
    return tfrecord_files


def _get_matching_s3_objects(
    bucket: str,
    prefix: Union[str, Tuple[str, ...]] = "",
    suffix: Union[str, Tuple[str, ...]] = "",
):
    """
    Generate objects in an S3 bucket.

    bucket: Name of the S3 bucket.
    prefix: Only fetch objects whose key starts with this prefix (optional).
    suffix: Only fetch objects whose keys end with this suffix (optional).
    """
    s3 = boto3.client("s3")
    paginator = s3.get_paginator("list_objects_v2")

    kwargs = {"Bucket": bucket}

    # We can pass the prefix directly to the S3 API.  If the user has passed
    # a tuple or list of prefixes, we go through them one by one.
    if isinstance(prefix, str):
        prefixes = (prefix,)
    else:
        prefixes = prefix

    for key_prefix in prefixes:
        kwargs["Prefix"] = key_prefix

        for page in paginator.paginate(**kwargs):
            try:
                contents = page["Contents"]
            except KeyError:
                break

            for obj in contents:
                key = obj["Key"]
                if key.endswith(suffix):
                    yield obj


def _get_matching_s3_keys(
    bucket: str,
    prefix: Union[str, Tuple[str, ...]] = "",
    suffix: Union[str, Tuple[str, ...]] = "",
):
    """
    Generate the keys in an S3 bucket.

    bucket: Name of the S3 bucket.
    prefix: Only fetch keys that start with this prefix (optional).
    suffix: Only fetch keys that end with this suffix (optional).
    """
    for obj in _get_matching_s3_objects(bucket, prefix, suffix):
        yield obj["Key"]


def _split_s3_url(url):
    """
    Source:
    https://stackoverflow.com/questions/42641315/s3-urls-get-bucket-name-and-path
    """
    parsed = urlparse(url, allow_fragments=False)
    bucket = parsed.netloc
    key = parsed.path.lstrip("/")
    if parsed.query:
        key += "?" + parsed.query
    return bucket, key



from .attention import EcaModule, attn_layer_factory  # noqa: F401
from .blurpool import BlurPool2D  # noqa: F401
from .classifier import ClassifierHead  # noqa: F401
from .conv import PadConv2D, PadDepthwiseConv2D, StdConv2D  # noqa: F401
from .drop import DropPath  # noqa: F401
from .factory import act_layer_factory, norm_layer_factory  # noqa: F401
from .initializers import FanoutInitializer  # noqa: F401
from .transformers import (  # noqa:F401
    MLP,
    ConvMLP,
    GatedMLP,
    GluMLP,
    PatchEmbeddings,
    interpolate_pos_embeddings,
    interpolate_pos_embeddings_grid,
)



"""
We provide an implementation and pretrained weights for the ConvNeXt models.

Paper: A ConvNet for the 2020s.
`[arXiv:2201.03545] <https://arxiv.org/abs/2201.03545>`_.

Original pytorch code and weights from
`Facebook Research <https://github.com/facebookresearch/ConvNeXt>`_.

This code has been ported from the
`timm <https://github.com/rwightman/pytorch-image-models>`_ implementation.

The following models are available.

* Models trained on ImageNet-1k

  * ``convnext_tiny``
  * ``convnext_small``
  * ``convnext_base``
  * ``convnext_large``

* Models trained on ImageNet-22k, fine-tuned on ImageNet-1k

  * ``convnext_tiny_in22ft1k``
  * ``convnext_small_in22ft1k``
  * ``convnext_base_in22ft1k``
  * ``convnext_large_in22ft1k``
  * ``convnext_xlarge_in22ft1k``

* Models trained on ImageNet-22k, fine-tuned on ImageNet-1k at 384 resolution

  * ``convnext_tiny_384_in22ft1k``
  * ``convnext_small_384_in22ft1k``
  * ``convnext_base_384_in22ft1k``
  * ``convnext_large_384_in22ft1k``
  * ``convnext_xlarge_384_in22ft1k``

* Models trained on ImageNet-22k

  * ``convnext_tiny_in22k``
  * ``convnext_small_in22k``
  * ``convnext_base_in22k``
  * ``convnext_large_in22k``
  * ``convnext_xlarge_in22k``
"""
# Copyright (c) Meta Platforms, Inc. and affiliates.
# All rights reserved.
# This source code is licensed under the MIT license
# Modifications and additions for timm by / Copyright 2022, Ross Wightman
# Copyright 2022 Marting Bruveris
from collections import OrderedDict
from dataclasses import dataclass
from typing import List, Tuple

import numpy as np
import tensorflow as tf

from tfimm.layers import MLP, ConvMLP, DropPath, norm_layer_factory
from tfimm.models import ModelConfig, keras_serializable, register_model
from tfimm.utils import IMAGENET_DEFAULT_MEAN, IMAGENET_DEFAULT_STD

# Model registry will add each entrypoint fn to this
__all__ = ["ConvNeXtConfig", "ConvNeXt"]


@dataclass
class ConvNeXtConfig(ModelConfig):
    """
    Configuration class for ConvNeXt models.

    Parameters:
        name: Name of the model.
        url: URL for pretrained weights.
        nb_classes: Number of classes for classification head.
        in_channels: Number of input image channels.
        input_size: Input image size (height, width)

        patch_size: Patchifying the image is implemented via a convolutional layer with
            kernel size and stride equal to ``patch_size``.
        embed_dim: Feature dimensions at each stage.
        nb_blocks: Number of blocks at each stage.
        mlp_ratio: Ratio of mlp hidden dim to embedding dim
        conv_mlp_block: There are two equivalent implementations of the ConvNeXt block,
            using either (1) 1x1 convolutions or (2) fully connected layers. In PyTorch
            option (2) also requires permuting channels, which is not needed in
            TensorFlow. We offer both implementations here, because some ``timm`` models
            use (1) while others use (2).

        drop_rate: Dropout rate.
        drop_path_rate: Dropout rate for stochastic depth.

        norm_layer: Normalization layer. See :func:`~norm_layer_factory` for possible
            values.
        act_layer: Activation function. See :func:`~act_layer_factory` for possible
            values.
        init_scale: Inital value for layer scale weights.

        crop_pct: Crop percentage for ImageNet evaluation.
        interpolation: Interpolation method for ImageNet evaluation.
        mean: Defines preprocessing function. If ``x`` is an image with pixel values
            in (0, 1), the preprocessing function is ``(x - mean) / std``.
        std: Defines preprpocessing function.

        first_conv: Name of first convolutional layer. Used by
            :func:`~tfimm.create_model` to adapt the number in input channels when
            loading pretrained weights.
        classifier: Name of classifier layer. Used by :func:`~tfimm.create_model` to
            adapt the classifier when loading pretrained weights.
    """

    nb_classes: int = 1000
    in_channels: int = 3
    input_size: Tuple[int, int] = (224, 224)
    patch_size: int = 4
    embed_dim: Tuple = (96, 192, 384, 768)
    nb_blocks: Tuple = (3, 3, 9, 3)
    mlp_ratio: float = 4.0
    conv_mlp_block: bool = False
    # Regularization
    drop_rate: float = 0.0
    drop_path_rate: float = 0.1
    # Other parameters
    norm_layer: str = "layer_norm_eps_1e-6"
    act_layer: str = "gelu"
    init_scale: float = 1e-6
    # Parameters for inference
    crop_pct: float = 0.875
    interpolation: str = "bicubic"
    # Preprocessing
    mean: Tuple[float, float, float] = IMAGENET_DEFAULT_MEAN
    std: Tuple[float, float, float] = IMAGENET_DEFAULT_STD
    # Weight transfer
    first_conv: str = "stem/0"
    classifier: str = "head/fc"


def _weight_initializers(seed=42):
    """Function returns initilializers to be used in the model."""
    kernel_initializer = tf.keras.initializers.TruncatedNormal(
        mean=0.0, stddev=0.02, seed=seed
    )
    bias_initializer = tf.keras.initializers.Zeros()
    return kernel_initializer, bias_initializer


class ConvNeXtBlock(tf.keras.layers.Layer):
    """
    ConvNeXt Block

    There are two equivalent implementations:
      (1) DwConv -> LayerNorm -> 1x1 Conv -> GELU -> 1x1 Conv
      (2) DwConv -> LayerNorm -> Linear -> GELU -> Linear;

    We offer both implementations, selected using the ``conv_mlp_block`` parameter,
    because the ``timm`` implementation also has both and uses (1) for some models
    and (2) for others. There is a slight performance difference in PyTorch, because
    (2) requires permuting axes.

    Args:
        dim (int): Number of input channels.
        drop_path (float): Stochastic depth rate. Default: 0.0
        ls_init_value (float): Init value for Layer Scale. Default: 1e-6.
    """

    def __init__(
        self,
        embed_dim: int,
        mlp_ratio: float,
        conv_mlp_block: bool,
        drop_rate: float,
        drop_path_rate: float,
        norm_layer: str,
        act_layer: str,
        init_scale: float,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.embed_dim = embed_dim
        self.mlp_ratio = mlp_ratio
        self.conv_mlp_block = conv_mlp_block
        self.drop_rate = drop_rate
        self.drop_path_rate = drop_path_rate
        self.norm_layer = norm_layer
        self.act_layer = act_layer
        self.init_scale = init_scale

        mlp_layer = ConvMLP if conv_mlp_block else MLP
        norm_layer = norm_layer_factory(norm_layer)
        kernel_initializer, bias_initializer = _weight_initializers()

        self.pad = tf.keras.layers.ZeroPadding2D(padding=3)
        self.conv_dw = tf.keras.layers.DepthwiseConv2D(
            kernel_size=7,
            depthwise_initializer=kernel_initializer,
            bias_initializer=bias_initializer,
            name="conv_dw",
        )
        self.norm = norm_layer(name="norm")
        self.mlp = mlp_layer(
            hidden_dim=int(mlp_ratio * embed_dim),
            embed_dim=embed_dim,
            drop_rate=drop_rate,
            act_layer=act_layer,
            kernel_initializer=kernel_initializer,
            bias_initializer=bias_initializer,
            name="mlp",
        )
        self.gamma = None
        self.drop_path = DropPath(drop_prob=drop_path_rate)

    def build(self, input_shape):
        self.gamma = self.add_weight(
            shape=(self.embed_dim,),
            initializer=tf.keras.initializers.Constant(value=self.init_scale),
            trainable=True,
            name="gamma",
        )

    def call(self, x, training=False):
        shortcut = x
        x = self.pad(x)
        x = self.conv_dw(x)
        x = self.norm(x, training=training)
        x = self.mlp(x, training=training)
        x = x * self.gamma
        x = self.drop_path(x, training=training)
        x = x + shortcut
        return x


class ConvNeXtStage(tf.keras.layers.Layer):
    """
    One stage of a ConvNeXt network: (optional) downsample layer, followed by a
    sequence of ``ConvNeXtBlox``s.
    """

    def __init__(
        self,
        stride: int,
        embed_dim: int,
        nb_blocks: int,
        mlp_ratio: float,
        conv_mlp_block: bool,
        drop_rate: float,
        drop_path_rate: np.ndarray,
        norm_layer: str,
        act_layer: str,
        init_scale: float,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.norm_layer = norm_layer

        norm_layer = norm_layer_factory(norm_layer)
        kernel_initializer, bias_initializer = _weight_initializers()

        if stride > 1:
            self.downsample_norm = norm_layer(name="downsample/0")
            self.downsample_conv = tf.keras.layers.Conv2D(
                filters=embed_dim,
                kernel_size=stride,
                strides=stride,
                kernel_initializer=kernel_initializer,
                bias_initializer=bias_initializer,
                name="downsample/1",
            )
        else:
            self.downsample_norm = None
            self.downsample_conv = None

        self.blocks = [
            ConvNeXtBlock(
                embed_dim=embed_dim,
                mlp_ratio=mlp_ratio,
                conv_mlp_block=conv_mlp_block,
                drop_rate=drop_rate,
                drop_path_rate=drop_path_rate[idx],
                norm_layer=self.norm_layer,
                act_layer=act_layer,
                init_scale=init_scale,
                name=f"blocks/{idx}",
            )
            for idx in range(nb_blocks)
        ]

    def call(self, x, training=False, return_features=False):
        features = OrderedDict()
        if self.downsample_conv is not None:
            x = self.downsample_norm(x, training=training)
            x = self.downsample_conv(x)
            features["downsample"] = x
        for idx, block in enumerate(self.blocks):
            x = block(x, training=training)
            features[f"block_{idx}"] = x
        return (x, features) if return_features else x


@keras_serializable
class ConvNeXt(tf.keras.Model):
    """
    Class implementing a ConvNeXt network.

    Paper: `A ConvNet for the 2020s <https://arxiv.org/pdf/2201.03545.pdf>`_.

    Parameters:
        cfg: Configuration class for the model.
        **kwargs: Arguments are passed to ``tf.keras.Model``.
    """

    cfg_class = ConvNeXtConfig

    def __init__(self, cfg: ConvNeXtConfig, **kwargs):
        kwargs["name"] = kwargs.get("name", cfg.name)
        super().__init__(**kwargs)
        self.cfg = cfg
        norm_layer = norm_layer_factory(cfg.norm_layer)
        kernel_initializer, bias_initializer = _weight_initializers()

        self.stem_conv = tf.keras.layers.Conv2D(
            filters=cfg.embed_dim[0],
            kernel_size=cfg.patch_size,
            strides=cfg.patch_size,
            kernel_initializer=kernel_initializer,
            bias_initializer=bias_initializer,
            name="stem/0",
        )
        self.stem_norm = norm_layer(name="stem/1")

        # Stochastic depth
        dpr = np.linspace(0.0, cfg.drop_path_rate, sum(cfg.nb_blocks))
        dpr = np.split(dpr, np.cumsum(cfg.nb_blocks))

        self.stages = []
        nb_stages = len(cfg.nb_blocks)
        for j in range(nb_stages):
            self.stages.append(
                ConvNeXtStage(
                    stride=2 if j > 0 else 1,
                    embed_dim=cfg.embed_dim[j],
                    nb_blocks=cfg.nb_blocks[j],
                    mlp_ratio=cfg.mlp_ratio,
                    conv_mlp_block=cfg.conv_mlp_block,
                    drop_rate=cfg.drop_rate,
                    drop_path_rate=dpr[j],
                    norm_layer=cfg.norm_layer,
                    act_layer=cfg.act_layer,
                    init_scale=cfg.init_scale,
                    name=f"stages/{j}",
                )
            )

        self.pool = tf.keras.layers.GlobalAveragePooling2D()
        self.norm = norm_layer(name="head/norm")
        self.flatten = tf.keras.layers.Flatten()
        self.drop = tf.keras.layers.Dropout(rate=cfg.drop_rate)
        self.fc = (
            tf.keras.layers.Dense(units=cfg.nb_classes, name="head/fc")
            if cfg.nb_classes > 0
            else tf.keras.layers.Activation("linear")  # Identity layer
        )

    @property
    def dummy_inputs(self) -> tf.Tensor:
        """Returns a tensor of the correct shape for inference."""
        return tf.zeros((1, *self.cfg.input_size, self.cfg.in_channels))

    @property
    def feature_names(self) -> List[str]:
        """
        Names of features, returned when calling ``call`` with ``return_features=True``.
        """
        _, features = self(self.dummy_inputs, return_features=True)
        return list(features.keys())

    def forward_features(
        self, x, training: bool = False, return_features: bool = False
    ):
        """
        Forward pass through model, excluding the classifier layer. This function is
        useful if the model is used as input for downstream tasks such as object
        detection.

        Arguments:
             x: Input to model
             training: Training or inference phase?
             return_features: If ``True``, we return not only the model output, but a
                dictionary with intermediate features.

        Returns:
            If ``return_features=True``, we return a tuple ``(y, features)``, where
            ``y`` is the model output and ``features`` is a dictionary with
            intermediate features.

            If ``return_features=False``, we return only ``y``.
        """
        features = OrderedDict()
        x = self.stem_conv(x)
        x = self.stem_norm(x, training=training)
        features["stem"] = x

        for stage_idx, stage in enumerate(self.stages):
            x = stage(x, training=training, return_features=return_features)
            if return_features:
                x, stage_features = x
                for key, val in stage_features.items():
                    features[f"stage_{stage_idx}/{key}"] = val
        features["conv_features"] = x

        return (x, features) if return_features else x

    def call(self, x, training: bool = False, return_features: bool = False):
        """
        Forward pass through the full model.

        Arguments:
             x: Input to model
             training: Training or inference phase?
             return_features: If ``True``, we return not only the model output, but a
                dictionary with intermediate features.

        Returns:
            If ``return_features=True``, we return a tuple ``(y, features)``, where
            ``y`` is the model output and ``features`` is a dictionary with
            intermediate features.

            If ``return_features=False``, we return only ``y``.
        """
        features = OrderedDict()
        x = self.forward_features(x, training, return_features)
        if return_features:
            x, features = x

        x = self.pool(x)
        x = self.norm(x, training=training)
        x = self.flatten(x)
        features["features"] = x
        x = self.drop(x, training=training)
        x = self.fc(x)
        features["logits"] = x
        return (x, features) if return_features else x


@register_model
def convnext_tiny():
    cfg = ConvNeXtConfig(
        name="convnext_tiny",
        url="[timm]",
        embed_dim=(96, 192, 384, 768),
        nb_blocks=(3, 3, 9, 3),
    )
    return ConvNeXt, cfg


@register_model
def convnext_small():
    cfg = ConvNeXtConfig(
        name="convnext_small",
        url="[timm]",
        embed_dim=(96, 192, 384, 768),
        nb_blocks=(3, 3, 27, 3),
    )
    return ConvNeXt, cfg


@register_model
def convnext_base():
    cfg = ConvNeXtConfig(
        name="convnext_base",
        url="[timm]",
        embed_dim=(128, 256, 512, 1024),
        nb_blocks=(3, 3, 27, 3),
    )
    return ConvNeXt, cfg


@register_model
def convnext_large():
    cfg = ConvNeXtConfig(
        name="convnext_large",
        url="[timm]",
        embed_dim=(192, 384, 768, 1536),
        nb_blocks=(3, 3, 27, 3),
    )
    return ConvNeXt, cfg


@register_model
def convnext_tiny_in22ft1k():
    cfg = ConvNeXtConfig(
        name="convnext_tiny_in22ft1k",
        url="[timm]",
        embed_dim=(96, 192, 384, 768),
        nb_blocks=(3, 3, 9, 3),
    )
    return ConvNeXt, cfg


@register_model
def convnext_small_in22ft1k():
    cfg = ConvNeXtConfig(
        name="convnext_small_in22ft1k",
        url="[timm]",
        embed_dim=(96, 192, 384, 768),
        nb_blocks=(3, 3, 27, 3),
    )
    return ConvNeXt, cfg


@register_model
def convnext_base_in22ft1k():
    cfg = ConvNeXtConfig(
        name="convnext_base_in22ft1k",
        url="[timm]",
        embed_dim=(128, 256, 512, 1024),
        nb_blocks=(3, 3, 27, 3),
    )
    return ConvNeXt, cfg


@register_model
def convnext_large_in22ft1k():
    cfg = ConvNeXtConfig(
        name="convnext_large_in22ft1k",
        url="[timm]",
        embed_dim=(192, 384, 768, 1536),
        nb_blocks=(3, 3, 27, 3),
    )
    return ConvNeXt, cfg


@register_model
def convnext_xlarge_in22ft1k():
    cfg = ConvNeXtConfig(
        name="convnext_xlarge_in22ft1k",
        url="[timm]",
        embed_dim=(256, 512, 1024, 2048),
        nb_blocks=(3, 3, 27, 3),
    )
    return ConvNeXt, cfg


@register_model
def convnext_tiny_384_in22ft1k():
    cfg = ConvNeXtConfig(
        name="convnext_tiny_384_in22ft1k",
        url="[timm]",
        input_size=(384, 384),
        embed_dim=(96, 192, 384, 768),
        nb_blocks=(3, 3, 9, 3),
    )
    return ConvNeXt, cfg


@register_model
def convnext_small_384_in22ft1k():
    cfg = ConvNeXtConfig(
        name="convnext_small_384_in22ft1k",
        url="[timm]",
        input_size=(384, 384),
        embed_dim=(96, 192, 384, 768),
        nb_blocks=(3, 3, 27, 3),
    )
    return ConvNeXt, cfg


@register_model
def convnext_base_384_in22ft1k():
    cfg = ConvNeXtConfig(
        name="convnext_base_384_in22ft1k",
        url="[timm]",
        input_size=(384, 384),
        embed_dim=(128, 256, 512, 1024),
        nb_blocks=(3, 3, 27, 3),
    )
    return ConvNeXt, cfg


@register_model
def convnext_large_384_in22ft1k():
    cfg = ConvNeXtConfig(
        name="convnext_large_384_in22ft1k",
        url="[timm]",
        input_size=(384, 384),
        embed_dim=(192, 384, 768, 1536),
        nb_blocks=(3, 3, 27, 3),
    )
    return ConvNeXt, cfg


@register_model
def convnext_xlarge_384_in22ft1k():
    cfg = ConvNeXtConfig(
        name="convnext_xlarge_384_in22ft1k",
        url="[timm]",
        input_size=(384, 384),
        embed_dim=(256, 512, 1024, 2048),
        nb_blocks=(3, 3, 27, 3),
    )
    return ConvNeXt, cfg


@register_model
def convnext_tiny_in22k():
    cfg = ConvNeXtConfig(
        name="convnext_tiny_in22k",
        url="[timm]",
        nb_classes=21841,
        embed_dim=(96, 192, 384, 768),
        nb_blocks=(3, 3, 9, 3),
    )
    return ConvNeXt, cfg


@register_model
def convnext_small_in22k():
    cfg = ConvNeXtConfig(
        name="convnext_small_in22k",
        url="[timm]",
        nb_classes=21841,
        embed_dim=(96, 192, 384, 768),
        nb_blocks=(3, 3, 27, 3),
    )
    return ConvNeXt, cfg


@register_model
def convnext_base_in22k():
    cfg = ConvNeXtConfig(
        name="convnext_base_in22k",
        url="[timm]",
        nb_classes=21841,
        embed_dim=(128, 256, 512, 1024),
        nb_blocks=(3, 3, 27, 3),
    )
    return ConvNeXt, cfg


@register_model
def convnext_large_in22k():
    cfg = ConvNeXtConfig(
        name="convnext_large_in22k",
        url="[timm]",
        nb_classes=21841,
        embed_dim=(192, 384, 768, 1536),
        nb_blocks=(3, 3, 27, 3),
    )
    return ConvNeXt, cfg


@register_model
def convnext_xlarge_in22k():
    cfg = ConvNeXtConfig(
        name="convnext_xlarge_in22k",
        url="[timm]",
        nb_classes=21841,
        embed_dim=(256, 512, 1024, 2048),
        nb_blocks=(3, 3, 27, 3),
    )
    return ConvNeXt, cfg



# Copyright 2018 The Google AI Language Team Authors and The HuggingFace Inc. team.
# Copyright (c) 2018, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import dataclasses
import functools

import tensorflow as tf


def keras_serializable(cls):
    """
    Decorate a Keras Layer class to support Keras serialization.

    This is done by:

    1. Adding a `cfg` dict to the Keras config dictionary in `get_config` (called by
       Keras at serialization time.
    2. Wrapping `__init__` to accept that `cfg` dict (passed by Keras at deserialization
       time) and convert it to a config object for the actual layer initializer.
    3. Registering the class as a custom object in Keras (if the Tensorflow version
       supports this), so that it does not need to be supplied in `custom_objects` in
       the call to `tf.keras.models.load_model`.

    Args:
        cls (a `tf.keras.layers.Layer` subclass):
            Typically the main network class in this project, in general must accept a
            `cfg` argument to its initializer.

    Returns:
        The same class object, with modifications for Keras serialization.
    """
    initializer = cls.__init__

    cfg_class = getattr(cls, "cfg_class", None)
    if cfg_class is None:
        raise AttributeError("Must set `cfg_class` to use @keras_serializable")

    @functools.wraps(initializer)
    def wrapped_init(self, *args, **kwargs):
        cfg = (
            args[0]
            if args and isinstance(args[0], cfg_class)
            else kwargs.pop("cfg", None)
        )

        if isinstance(cfg, dict):
            cfg = cfg_class(**cfg)
            initializer(self, cfg, *args, **kwargs)
        elif isinstance(cfg, cfg_class):
            if len(args) > 0:
                initializer(self, *args, **kwargs)
            else:
                initializer(self, cfg, *args, **kwargs)
        else:
            raise ValueError("Must pass either `cfg` (ModelConfig) or `cfg` (dict)")

        # This is probably just a safety net
        self._cfg = cfg
        self._kwargs = kwargs

    cls.__init__ = wrapped_init

    def get_config(self):
        return dataclasses.asdict(self.cfg)

    cls.get_config = get_config

    @classmethod
    def from_config(_cls, _cfg):
        _cfg = cfg_class(**_cfg)
        return _cls(_cfg)

    cls.from_config = from_config

    cls._keras_serializable = True
    if hasattr(tf.keras.utils, "register_keras_serializable"):
        cls = tf.keras.utils.register_keras_serializable()(cls)
    return cls



from dataclasses import dataclass

from .registry import cfg_serializable


@cfg_serializable
@dataclass
class Timekeeping:
    """
    The purpose of this class is to coordinate "time" across the trainer class, the
    problem class and the optimizer. By "time" we mean:
     - How long are we going to train for (`nb_epochs`)?
     - How quickly is time passing (`batch_size`)?
     - How long is one epoch (`nb_samples_per_epoch`)?
    This class allows the user to define these fundamental objects in one place in the
    config.
    """

    # How long should the training last?
    nb_epochs: int
    # How many samples do we see per training step, i.e., per call to `train_step()`?
    # We use this value only for time keeping, the dataset is free to return batches
    # of arbitrary shape.
    batch_size: int
    # The trainer class will increase the epoch counter when (a) the `train_ds`
    # iterator is exhausted or (b) we have seen `nb_samples_per_epoch. When
    # `train_ds` is a finite dataset and `nb_samples_per_epoch > 0`, then both
    # conditions will trigger a new epoch. This can lead to epochs of uneven length.
    # Use at your own risk.
    #
    # On the other hand, `nb_samples_per_epoch` will be used by various learning
    # rate schedulers to know when to change the learning rate.
    # If `nb_samples_per_epoch = -1`, most learning rate schedulers will not work
    #
    # The recommendation is to use infinite datasets, e.g., via `ds.repeat()` combined
    # with a value for `nb_samples_per_epoch`.
    nb_samples_per_epoch: int = -1

    @property
    def nb_steps(self) -> int:
        """Total number of steps for the training."""
        if self.nb_samples_per_epoch == -1:
            raise ValueError(
                "Cannot compute `nb_steps`, if `nb_samples_per_epoch` is not set."
            )
        return self.nb_epochs * self.nb_steps_per_epoch

    @property
    def nb_steps_per_epoch(self) -> int:
        """Number of training steps in each epoch."""
        if self.nb_samples_per_epoch == -1:
            raise ValueError(
                "Cannot compute `nb_steps_per_epoch`, if `nb_samples_per_epoch` is not "
                "set."
            )
        return self.nb_samples_per_epoch // self.batch_size



import logging
from dataclasses import dataclass

import tensorflow as tf

try:
    import tensorflow_datasets as tfds
except ImportError:
    tfds = None
    logging.info("Could not import `tensorflow_datasets`.")

from ..registry import cfg_serializable


@dataclass
class TFDSConfig:
    dataset_name: str
    split: str
    input_size: tuple
    batch_size: int
    include_label: bool = True
    repeat: bool = False
    shuffle: bool = False
    nb_samples: int = None
    cache: bool = False
    dtype: str = "float32"


@cfg_serializable
class TFDSWrapper:
    """Wrapper class around tensorflow datasets."""

    cfg_class = TFDSConfig

    def __init__(self, cfg: TFDSConfig):
        self.cfg = cfg
        self.root_ds = tfds.load(
            self.cfg.dataset_name,
            split=self.cfg.split,
            shuffle_files=self.cfg.shuffle,
            as_supervised=True,
            # We disable auto-caching at this point, because if we use `nb_samples`,
            # TF will print repeated warnings that we are not exhausting the iterator
            # of a cached dataset. Use `cfg.cache` instead, which caches the dataset
            # after applying `nb_samples`.
            read_config=tfds.ReadConfig(try_autocache=False),
        )

    def get_ds(self):
        ds = self.root_ds
        if self.cfg.nb_samples:
            ds = ds.take(self.cfg.nb_samples)
        if self.cfg.cache:
            ds = ds.cache()
        if self.cfg.shuffle:
            ds = ds.shuffle(buffer_size=3000)
        if self.cfg.repeat:
            ds = ds.repeat()
        if self.cfg.batch_size:
            ds = ds.batch(self.cfg.batch_size, drop_remainder=False)
        ds = ds.map(self.decode_batch, num_parallel_calls=tf.data.experimental.AUTOTUNE)
        ds = ds.prefetch(tf.data.experimental.AUTOTUNE)
        return ds

    def decode_batch(self, img, label):
        if self.cfg.input_size:
            img = tf.image.resize(img, size=self.cfg.input_size)
        if self.cfg.include_label:
            return img, label
        else:
            return img

    def __iter__(self):
        return iter(self.get_ds())
